/***********************************************************************

	File:	sthgp_indicators.h
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	Compute strength indicators for polytope STHGP(n) via
	closed-form formulae.

************************************************************************

	Modification Log:

	a-1:	11/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#ifndef STHGP_INDICATORS_H
#define	STHGP_INDICATORS_H

#include <gmp.h>

/*
 * Return the number of extreme points of STHGP(n).
 */

extern void	sthgp_num_extreme_points (mpz_ptr res, int n);

/*
 * Let e \in E such that |e| = k.  Return the number of STHGP(n)
 * extreme points that contain edge e.
 */

extern void	sthgp_num_edge_extreme_points (mpz_ptr res, int n, int k);

/*
 * Return the number of STHGP(n) extreme points incident to the
 * non-negativity inequality for an edge e of cardinality k.
 */

extern void	sthgp_nonneg_ep_count (mpz_ptr		res,
				       int		n,
				       int		k);

/*
 * Return the Extreme Point Ratio for an STHGP(n) non-negativity inequality
 * for an edge e of cardinality k.
 */

extern void	sthgp_nonneg_epr (mpq_ptr	res,
				  int		n,
				  int		k);

/*
 * Return the Centroid Distance squared for an STHGP(n) non-negativity
 * inequality for an edge e of cardinality k.
 */

extern void	sthgp_nonneg_cd2 (mpq_ptr	res,
				  int		n,
				  int		k);

/*
 * Return the number of STHGP(n) extreme points incident to a
 * subtour inequality S of cardinality k.
 */

extern void	sthgp_subtour_ep_count (mpz_ptr		res,
					int		n,
					int		k);

/*
 * Return the number of STHGP(n) extreme points incident to a
 * subtour inequality S of cardinality k.  This uses an "alternate"
 * formula that is much more computationally efficient for large n.
 * This formula is not "closed-form," however, because it contains
 * non-standard functions defined only via non-standard recurrence
 * relations.
 */

extern void	sthgp_subtour_alternate_ep_count (mpz_ptr	res,
						  int		n,
						  int		k);

/*
 * Return the Extreme Point Ratio for an STHGP(n) subtour S of cardinality k.
 */

extern void	sthgp_subtour_epr (mpq_ptr	res,
				   int		n,
				   int		k);

/*
 * Set array[k] to the number of STHGP(n) extreme points incident
 * to a subtour inequality S of cardinality k (for all 2 <= k <= n).
 * This means that array[n] is the total number of extreme points.
 */

extern void	sthgp_all_subtour_ep_count (mpz_ptr array, int n);

/*
 * Return the Centroid Distance squared for an STHGP(n) subtour S of
 * cardinality k.
 */

extern void	sthgp_subtour_cd2 (mpq_ptr	res,
				   int		n,
				   int		k);

/*
 * Set array[k] to the Centroid Distance squared for an STHGP(n) subtour
 * S of cardinality k (for all 2 <= k <= n).
 */

extern void	sthgp_all_subtour_cd2 (mpq_ptr array, int n);

/*
 * Compute t(n,k).
 */

extern void	sthgp_subtour_cd2_t (mpz_ptr res, int n, int k);

/*
 * Compute Alpha(n).
 */

extern void	sthgp_subtour_cd2_Alpha (mpz_ptr res, int n);

/*
 * Compute Beta(n,k).
 */

extern void	sthgp_subtour_cd2_Beta (mpz_ptr res, int n, int k);

/*
 * Compute Gamma(n,k).
 */

extern void	sthgp_subtour_cd2_Gamma (mpz_ptr res, int n, int k);

/*
 * Shut down the sthgp_indicators module, freeing all memory it
 * statically holds.
 */

extern void	sthgp_indicators_shutdown (void);

#endif
