/***********************************************************************

	$Id: pmoment.c,v 1.1 2025/01/26 02:31:43 warme Exp $

	File:	pmoment.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

************************************************************************

	Poisson moments.

************************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "pmoment.h"

#include "fatal.h"
#include <gmp.h>
#include "gmpmisc.h"
#include "logic.h"
#include "stirling2.h"
#include "zmemo.h"

/*
 * Local Functions
 */

static void	compute_pmoment (mpz_ptr res, int n, int k);

/*
 * Local Data
 */

static struct ZMemo *	memo;

/*
 * Return the moment E[X^k], where X is a Poisson random variable
 * having mean of n.
 */

	void
pmoment (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Mean of Poisson random variable */
int		k		/* IN: Which moment to calculate */
)
{
struct ZMemo *	mp1;
struct ZMemo *	mp2;
mpz_ptr		zp;
mpz_t		TEMP;

	mp1 = memo;
	if (mp1 EQ NULL) {
		mp1 = zmemo_create_mrow ();
		memo = mp1;
	}

	mp2 = zmemo_get_mrow (mp1, n);
	if (mp2 EQ NULL) {
		mp2 = zmemo_create_zrow ();
		zmemo_set_mrow (mp1, n, mp2);
	}

	zp = zmemo_get_zrow (mp2, k);
	if (zp NE NULL) {
		mpz_set (res, zp);
	}
	else {
		mpz_init (TEMP);
		compute_pmoment (TEMP, n, k);

		zmemo_set_zrow (mp2, k, TEMP);

		mpz_set (res, TEMP);

		mpz_clear (TEMP);
	}
}

/*
 * Return the moment E[X^k], where X is a Poisson random variable
 * having mean of n.
 */

	static
	void
compute_pmoment (

mpz_ptr		res,		/* IN/OUT: Result */
int		n,		/* IN: Mean of Poisson random variable */
int		k		/* IN: Which moment to calculate */
)
{
int		i;
mpz_t		SUM, TERM, POW;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 0);
	FATAL_ERROR_IF (k < 0);

	mpz_init (SUM);
	mpz_init (TERM);
	mpz_init (POW);

	mpz_set_ui (POW, 1);
	for (i = 0; i <= k; i++) {
		stirling2 (TERM, k, i);
		mpz_mul (TERM, TERM, POW);
		mpz_add (SUM, SUM, TERM);

		mpz_mul_ui (POW, POW, n);
	}
	mpz_set (res, SUM);

	mpz_clear (POW);
	mpz_clear (TERM);
	mpz_clear (SUM);
}

/*
 * Shut down pmoment module, freeing all statically-held memory.
 */

	void
pmoment_shutdown (void)

{
	stirling2_shutdown ();

	zmemo_free (memo);
	memo = NULL;
}
