/***********************************************************************

	File:	sthgp_polytope.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	A wrapper for polytope STHGP(n).

************************************************************************

	Modification Log:

	a-1:	09/07/2015	warme
		: Created.
	a-1:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "sthgp_polytope.h"

#include "bitmaskmacros.h"
#include "fatal.h"
#include <limits.h>
#include "logic.h"
#include <math.h>
#include "new.h"
#include "polytope.h"
#include "s3itypes.h"
#include "scanutils.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Global Routines
 */

struct Polytope *	make_STHGP_polytope (int n);

/*
 * The actual implementation of the STHGP(n) Polytope.
 * This is a "derived class" from Polytope.
 */

struct STHGP {
	struct Polytope	base;		/* The base class instance */
	int		nverts;		/* Number of vertices */
	int		nedges;		/* Number of hyperedges */
	int **		edge_list;	/* To access all vertices of edge e: */
					/* ep1 = edge_list [e];		*/
					/* ep2 = edge_list [e+1];	*/
					/* while (ep1 < ep2) {		*/
					/*	v = *ep1++;		*/
					/* }				*/
};

struct gen {
	int		nverts;		/* Number of vertices */
	int		count;		/* Number of edges generated */
	int **		edge_list;	/* Array of edge vertex lists */
	int *		genp;		/* output buffer pointer */
	int *		stack;		/* base of stack */
};

/*
 * The stack-oriented data structures for enumerating hypertrees.
 */

struct sinfo {
	int			nmasks;
	int			kmasks;
	int *			fsets;
	int **			edge_list;
	bitmap_t *		solution;
	bitmap_t *		terms_left;
	bitmap_t *		compat;
	bitmap_t *		incompat;
	bitmap_t *		try;
	bitmap_t *		omit;
	bitmap_t *		cmasks;
	bitmap_t *		incmasks;
	bitmap_t *		edge_tmap;
	int8u **		one_bits_in_byte;
	int **			term_trees;
	int			ntrees;		/* Number of trees generated */
	struct iv_mask *	root;		/* Linked list of trees */
	struct iv_mask **	hookp;		/* For appending to root */
};

/*
 * Linked list of generated incidence vectors (bitmask form).
 */

struct iv_mask {
	bitmap_t *		mask;	/* Mask of edges in the tree */
	struct iv_mask *	next;	/* Next incident vector in list */
};

/*
 * Local Routines
 */

static int *		compute_binomial_row (int n);
static const char **	create_edge_names (int		nverts,
					   int		nedges,
					   int **	edge_list);
static int8u **		create_one_bits_in_byte_table (void);
static struct IVecSet *	enumerate_hypertrees (const struct Polytope * poly);
static void		free_STHGP (struct Polytope * poly);
static void		gen_h (int, int, int *, struct gen *);
static int **		generate_hyperedges (int nverts, int nedges);
static struct Constraint *
			get_STHGP_constraint (const struct Polytope *	_poly,
					      const char *		name);
static struct Constraint *
			get_STHGP_nonneg_constraint (
					const struct STHGP *	poly,
					const char *		s);
static struct Constraint *
			get_STHGP_subtour (const struct STHGP *	poly,
					   const char *		s);
static void		ht_recurse (struct sinfo * sip, int level, int nleft);

/*
 * Return an object representing the polytope STHGP(n).
 */

	struct Polytope *
make_STHGP_polytope (

int		nverts		/* IN: number of vertices in complete graph */
				/*     from which tours are taken */
)
{
int			j, k;
int			nedges, maxlen;
struct STHGP *		poly;
int **			edge_list;
const char **		edge_names;
char			buffer [32];

	if (nverts < 2) {
		fprintf (stderr, "STHGP(n) requires n >= 2.\n");
		return (NULL);
	}

	poly = NEW (struct STHGP);
	memset (poly, 0, sizeof (*poly));

	nedges	= (1 << nverts) - nverts - 1;

	edge_list = generate_hyperedges (nverts, nedges);

	edge_names = create_edge_names (nverts, nedges, edge_list);

	sprintf (buffer, "STHGP(%d)", nverts);

	maxlen = 0;
	for (j = 0; j < nedges; j++) {
		k = strlen (edge_names [j]);
		if (k > maxlen) {
			maxlen = k;
		}
	}

	poly -> base.name	= strdup (buffer);
	poly ->	base.dim	= nedges;
	poly -> base.ename	= edge_names;
	poly -> base.maxlen	= maxlen;
	poly -> base.get_extreme_points		= enumerate_hypertrees;
	poly -> base.get_constraint		= get_STHGP_constraint;
	poly -> base.free			= free_STHGP;

	poly -> nverts		= nverts;
	poly -> nedges		= nedges;
	poly -> edge_list	= edge_list;

	return (&(poly -> base));
}

/*
 * This routine generates all of the hyperedges of size 2 or more.
 */

	static
	int **
generate_hyperedges (

int			nverts,		/* IN: number of vertices */
int			nedges		/* IN: number of hyperedges */
)
{
int		i, k;
int *		ip;
int **		edge_list;
int *		binomial_row;
struct gen	gen;

	/* Compute sum of sizes of all hyperedges. */
	binomial_row = compute_binomial_row (nverts);
	k = 0;
	for (i = 2; i <= nverts; i++) {
		k += i * binomial_row [i];
	}

	edge_list = NEWA (nedges + 1, int *);
	ip	  = NEWA (k, int);

	gen.nverts		= nverts;
	gen.count		= 0;
	gen.edge_list		= edge_list;
	gen.genp		= ip;
	gen.stack		= NEWA (nverts, int);

	for (i = 2; i <= nverts; i++) {
		gen_h (0, i, gen.stack, &gen);
	}
	free (gen.stack);

	FATAL_ERROR_IF (ip + k NE gen.genp);
	FATAL_ERROR_IF (gen.count NE nedges);

	edge_list [nedges] = gen.genp;

	free (binomial_row);

	return (edge_list);
}

/*
 * Recursive portion of hyperedge generator.
 */

	static
	void
gen_h (

int		v,		/* IN - current vertex number */
int		nleft,		/* IN - number of vertices left to choose */
int *		sp,		/* IN - current stack pointer */
struct gen *	gp		/* IN/OUT - generation info */
)
{
int *		ip;

	if (nleft <= 0) {
		/* We have a complete hyperedge! */
		gp -> edge_list [gp -> count] = gp -> genp;
		++(gp -> count);
		for (ip = gp -> stack; ip < sp; ip++) {
			*(gp -> genp)++ = *ip;
		}
		return;
	}

	if (v + nleft <= gp -> nverts) {
		/* Include vertex v. */
		*sp = v;
		gen_h (v + 1, nleft - 1, sp + 1, gp);

		/* Exclude vertex v. */
		gen_h (v + 1, nleft, sp, gp);
	}
}

/*
 * Compute the Nth row of the binomial coefficients, via the
 * recurrence relation.
 */

	static
	int *
compute_binomial_row (

int		n	/* IN: which row to generate */
)
{
int		i;
int		j;
int *		prev;
int *		cur;
int *		temp;

	prev = NEWA (n + 1, int);
	cur  = NEWA (n + 1, int);

	/* {0 \choose 0} ==> 1 */
	j = 0;
	cur [0] = 1;

	for (; j <= n; j++) {
		temp = cur;
		cur  = prev;
		prev = temp;

		/* {n \choose 0} ==> 1 */
		cur [0] = 1;
		for (i = 1; i < j; i++) {
			cur [i] = prev [i - 1] + prev [i];
		}
		cur [j] = 1;
	}

	free (prev);

	return (cur);
}

/*
 * Create names for each of the edges.
 */

	static
	const char **
create_edge_names (

int		nverts,		/* IN: number of vertices */
int		nedges,		/* IN: number of edges */
int **		edge_list	/* IN: vertices contained in each edge */
)
{
int		j, v;
const char **	edge_names;
char *		s;
int *		ip1;
int *		ip2;
char		buffer [128];

	edge_names = NEWA (nedges, const char *);

	if (nverts <= 10) {
		/* We can use single digits for each vertex. */
		for (j = 0; j < nedges; j++) {
			s = &buffer [0];
			*s++ = 'x';
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				*s++ = '0' + v;
			}
			*s = '\0';
			edge_names [j] = strdup (buffer);
		}
	}
	else {
		/* At least one vertex requires more than 1 digit. */
		for (j = 0; j < nedges; j++) {
			s = &buffer [0];
			*s++ = 'x';
			ip1 = edge_list [j];
			ip2 = edge_list [j + 1];
			while (ip1 < ip2) {
				v = *ip1++;
				sprintf (s, "_%d", v);
				s = strchr (s, '\0');
			}
			edge_names [j] = strdup (buffer);
		}
	}
	return (edge_names);
}

/*
 * Enumerate all hypertrees.
 */

	static
	struct IVecSet *
enumerate_hypertrees (

const struct Polytope *	_poly		/* IN: polytope to enumerate	*/
					/*     all hypertrees for	*/
)
{
int			i, j, k, t;
int			nverts, nedges;
int			kmasks, nmasks;
int			count;
int			edges_per_term;
size_t			size, csize;
const struct STHGP *	poly;
int8u **		one_bits_in_byte;
int **			edge_list;
bitmap_t *		edge_tmap;
int **			term_trees;
bitmap_t *		cmasks;
bitmap_t *		incmasks;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
bitmap_t *		array;
struct iv_mask *	list;
int *			ip;
int *			tp1;
int *			endp;
int **			ptr;
struct IVecSet *	xpts;
struct sinfo		sinfo;

	FATAL_ERROR_IF (_poly EQ NULL);
	FATAL_ERROR_IF (_poly -> get_extreme_points NE enumerate_hypertrees);

	poly = (const struct STHGP *) _poly;

	nverts		= poly -> nverts;
	nedges		= poly -> nedges;
	edge_list	= poly -> edge_list;

	FATAL_ERROR_IF (poly -> base.dim NE nedges);

	nmasks = BMAP_ELTS (nedges);
	kmasks = BMAP_ELTS (nverts);

	init_nbits ();
	one_bits_in_byte = create_one_bits_in_byte_table ();

	/* Build edge_tmap -- a bitmap of vertices for each edge. */
	size = nedges * kmasks;
	edge_tmap = NEWA (size, bitmap_t);
	(void) memset (edge_tmap, 0, size * sizeof (bitmap_t));
	bp1 = edge_tmap;
	for (i = 0; i < nedges; i++) {
		ip = edge_list [i];
		endp = edge_list [i + 1];
		while (ip < endp) {
			j = *ip++;
			SETBIT (bp1, j);
		}
		bp1 += kmasks;
	}

	/* Build term_trees -- a list of hyperedges for each terminal. */
	edges_per_term = (1 << (nverts - 1)) - 1;
	ip = NEWA (nverts * edges_per_term, int);
	term_trees	= NEWA (nverts + 1, int *);
	ptr		= NEWA (nverts, int *);
	for (i = 0; i < nverts; i++) {
		term_trees [i]	= ip;
		ptr [i]		= ip;
		ip += edges_per_term;
	}
	term_trees [i] = ip;
	for (i = 0; i < nedges; i++) {
		ip = edge_list [i];
		endp = edge_list [i + 1];
		while (ip < endp) {
			j = *ip++;
			*(ptr [j])++ = i;
		}
	}
	for (i = 0; i < nverts; i++) {
		FATAL_ERROR_IF (ptr [i] NE term_trees [i + 1]);
	}
	free (ptr);

	/* Build cmasks -- a mask of compatible edges for each edge. */
	/* Build incmasks -- a mask of incompatible edges for each edge. */
	size = nedges * nmasks;
	cmasks	 = NEWA (size, bitmap_t);
	incmasks = NEWA (size, bitmap_t);
	(void) memset (cmasks, 0, size * sizeof (bitmap_t));
	(void) memset (incmasks, 0, size * sizeof (bitmap_t));
	for (i = 0; i < nedges; i++) {
		SETBIT (&incmasks [i * nmasks], i);
		for (j = i + 1; j < nedges; j++) {
			count = 0;
			ip = edge_list [i];
			endp = edge_list [i + 1];
			bp1 = &edge_tmap [j * kmasks];
			while (ip < endp) {
				k = *ip++;
				if (BITON (bp1, k)) {
					++count;
				}
			}
			if (count EQ 1) {
				SETBIT (&cmasks [i * nmasks], j);
				SETBIT (&cmasks [j * nmasks], i);
			}
			else if (count > 1) {
				SETBIT (&incmasks [i * nmasks], j);
				SETBIT (&incmasks [j * nmasks], i);
			}
		}
	}

	memset (&sinfo, 0, sizeof (sinfo));

	csize = nverts + 1;
	csize *= nmasks;

	sinfo.nmasks		= nmasks;
	sinfo.kmasks		= kmasks;
	sinfo.fsets		= NEWA (nverts, int);
	sinfo.edge_list		= edge_list;
	sinfo.solution		= NEWA (nmasks, bitmap_t);
	sinfo.terms_left	= NEWA ((nverts + 1) * kmasks, bitmap_t);
	sinfo.compat		= NEWA (csize, bitmap_t);
	sinfo.incompat		= NEWA (csize, bitmap_t);
	sinfo.try		= NEWA (csize, bitmap_t);
	sinfo.omit		= NEWA (csize, bitmap_t);
	sinfo.cmasks		= cmasks;
	sinfo.incmasks		= incmasks;
	sinfo.edge_tmap		= edge_tmap;
	sinfo.one_bits_in_byte	= one_bits_in_byte;
	sinfo.term_trees	= term_trees;
	sinfo.ntrees		= 0;
	sinfo.root		= NULL;
	sinfo.hookp		= &sinfo.root;

	bp1 = sinfo.terms_left;
	for (i = 0; i < kmasks; i++) {
		*bp1++ = 0;
	}
	for (i = 0; i < nverts; i++) {
		SETBIT (sinfo.terms_left, i);
	}

	size = (nmasks * sizeof (bitmap_t));
	(void) memset (sinfo.compat, 0, size);
	(void) memset (sinfo.incompat, 0, size);
	(void) memset (sinfo.try, 0, size);
	(void) memset (sinfo.omit, 0, size);
	(void) memset (sinfo.solution, 0, size);

	/* We start at terminal 0... */

	tp1	= term_trees [0];
	endp	= term_trees [1];
	while (tp1 < endp) {
		t = *tp1++;

		sinfo.fsets [0] = t;
		SETBIT (sinfo.solution, t);

		bp1 = sinfo.terms_left;
		bp2 = &edge_tmap [t * kmasks];
		bp3 = bp1 + kmasks;
		for (i = 0; i < kmasks; i++) {
			*bp3++ = *bp1++ & ~*bp2++;
		}

		bp1 = &cmasks [t * nmasks];
		bp2 = sinfo.compat + nmasks;
		for (i = 0; i < nmasks; i++) {
			*bp2++ = *bp1++;
		}

		bp1 = &incmasks [t * nmasks];
		bp2 = sinfo.omit;
		bp3 = sinfo.incompat + nmasks;
		for (i = 0; i < nmasks; i++) {
			*bp3++ = *bp1++ | *bp2++;
		}

		ht_recurse (&sinfo,
			    1,
			    nverts - (edge_list [t + 1] - edge_list [t]));

		CLRBIT (sinfo.solution, t);
		SETBIT (sinfo.omit, t);
	}

	/* Copy the incidence vectors into a single array of bit masks.	*/
	/* Each incidence vector is nmasks words long.			*/

	array = NEWA (sinfo.ntrees * nmasks, bitmap_t);
	bp1 = array;
	for (;;) {
		list = sinfo.root;
		if (list EQ NULL) break;

		sinfo.root = list -> next;
		list -> next = NULL;

		memcpy (bp1, list -> mask, nmasks * sizeof (bitmap_t));
		bp1 += nmasks;

		free (list -> mask);
		free (list);
	}

	xpts = NEW (struct IVecSet);

	xpts -> nvecs	= sinfo.ntrees;
	xpts -> dim	= nedges;
	xpts -> nmasks	= nmasks;
	xpts -> array	= array;

	free (sinfo.fsets);
	sinfo.edge_list = NULL;
	free (sinfo.solution);
	free (sinfo.terms_left);
	free (sinfo.compat);
	free (sinfo.incompat);
	free (sinfo.try);
	free (sinfo.omit);
	free (incmasks);
	free (cmasks);

	free (edge_tmap);
	free (one_bits_in_byte [0]);
	free (one_bits_in_byte);

	free (term_trees [0]);
	free (term_trees);

	return (xpts);
}

/*
 * This routine initializes a table that is used to rapidly convert an
 * 8-bit byte into a sequence of bit-positions that are on within that
 * byte.  For example, it translates the byte 0xB7 into {0, 1, 2, 4, 5, 7}.
 */

	static
	int8u **
create_one_bits_in_byte_table (void)

{
int		i;
int		j;
int		byte;
int8u *		byte_array;
int8u **	ptr_array;
int8u *		p;

	byte_array = NEWA (1024, int8u);
	ptr_array	= NEWA (256 + 1, int8u *);

	p = &byte_array [0];
	for (i = 0; i < 256; i++) {
		ptr_array [i] = p;
		byte = i;
		j = 0;
		while (byte > 0) {
			if ((byte & 0x01) NE 0) {
				*p++ = j;
			}
			++j;
			byte >>= 1;
		}
	}

	/* Terminate last table entry. */
	ptr_array [i] = p;

	return (ptr_array);
}

/*
 * This routine performs the recursive portion of the backtrack search.
 */

	static
	void
ht_recurse (

struct sinfo *	sip,		/* IN/OUT - search info */
int		level,		/* IN - recursion level */
int		nleft		/* IN - number of terminals left to connect */
)
{
int			i, j, k;
int			nmasks, kmasks;
int			fset_offset;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
bitmap_t *		bp4;
bitmap_t		mask;
bitmap_t *		terms_left;
bitmap_t *		compat;
bitmap_t *		incompat;
bitmap_t *		try;
bitmap_t *		omit;
bitmap_t *		cmasks;
bitmap_t *		incmasks;
int			i1;
int **			edge_list;
int8u **		one_bits_in_byte;
bitmap_t *		edge_tmap;
bitmap_t *		wordp;
bitmap_t		word;
int			basebit;
int			bitno;
int8u *			p1;
int8u *			p2;
struct iv_mask *	list;
int			byte;

	nmasks	= sip -> nmasks;
	kmasks	= sip -> kmasks;

	if (nleft <= 0) {
		FATAL_ERROR_IF (nleft < 0);

		/* All terminals have been connected!  We have	*/
		/* a new hypertree.  Add it to the list.	*/

		bp1 = NEWA (nmasks, bitmap_t);
		memcpy (bp1, sip -> solution, nmasks * sizeof (bitmap_t));

		list = NEW (struct iv_mask);
		list -> mask	= bp1;
		list -> next	= NULL;
		*(sip -> hookp) = list;
		sip -> hookp = &(list -> next);

		++(sip -> ntrees);

		return;
	}

	terms_left	= sip -> terms_left + level * kmasks;

	fset_offset	= level * nmasks;

	compat		= sip -> compat	  + fset_offset;
	incompat	= sip -> incompat + fset_offset;
	try		= sip -> try	  + fset_offset;
	omit		= sip -> omit	  + fset_offset;

	/* Determine the set of trees we will try to add to the	*/
	/* partial solution.					*/

	bp1 = compat;
	bp2 = incompat;
	bp3 = try;
	for (i = 0; i < nmasks; i++) {
		*bp3++ = *bp1++ & ~*bp2++;
	}

	bp1 = omit;
	for (i = 0; i < nmasks; i++) {
		*bp1++ = 0;
	}

	/* Loop over each feasible hyperedge. */
	edge_list	 = sip -> edge_list;
	one_bits_in_byte = sip -> one_bits_in_byte;
	edge_tmap	 = sip -> edge_tmap;
	cmasks		 = sip -> cmasks;
	incmasks	 = sip -> incmasks;
	wordp = try;
	for (j = 0, basebit = 0; j < nmasks; basebit += BPW, j++) {
		word = *wordp++;
		if (word EQ 0) continue;
		for (bitno = basebit; word NE 0; bitno += 8, word >>= 8) {
			byte = (word & 0xFF);
			if (byte EQ 0) continue;
			p1 = one_bits_in_byte [byte];
			p2 = one_bits_in_byte [byte + 1];
			while (p1 < p2) {
				k = bitno + *p1++;
				/* Hyperedge K is feasible.  Test to see */
				/* if it is truly compatible...		 */
				bp1 = &edge_tmap [k * kmasks];
				bp2 = terms_left;
				i1 = 0;
				for (i = 0; i < kmasks; i++) {
					mask = *bp1++ & ~*bp2++;
					i1 += NBITSON (mask);
				}
				if (i1 NE 1) continue;
				/* Tree K intersects the current partial */
				/* solution at exactly one point!	 */

				bp1 = compat;
				bp2 = &cmasks [k * nmasks];
				bp3 = bp1 + nmasks;
				for (i = 0; i < nmasks; i++) {
					*bp3++ = *bp1++ | *bp2++;
				}

				bp1 = incompat;
				bp2 = &incmasks [k * nmasks];
				bp3 = omit;
				bp4 = bp1 + nmasks;
				for (i = 0; i < nmasks; i++) {
					*bp4++ = (*bp1++ | *bp2++ | *bp3++);
				}

				bp1 = terms_left;
				bp2 = &edge_tmap [k * kmasks];
				bp3 = bp1 + kmasks;
				for (i = 0; i < kmasks; i++) {
					*bp3++ = *bp1++ & ~*bp2++;
				}

				sip -> fsets [level] = k;
				SETBIT (sip -> solution, k);

				ht_recurse (sip,
					    level + 1,
					    nleft
					      - (edge_list [k + 1]
						 - edge_list [k]) + 1);

				CLRBIT (sip -> solution, k);
				SETBIT (omit, k);
			}
		}
	}
}

/*
 * Decode the given name and return the corresponding inequality.
 */

	static
	struct Constraint *
get_STHGP_constraint (

const struct Polytope *	_poly,		/* IN: polytope to get constraint of */
const char *		name		/* IN: name of constraint */
)
{
const struct STHGP *	poly;
const char *		s;
struct Constraint *	cp;

	static const char *	sec_names [] = {
		"sec-", "sec", "subtour-", "subtour", NULL
	};
	static const char *	nonneg_names [] = {
		"nn-", "nn", "nonneg-", "nonneg", NULL
	};

	FATAL_ERROR_IF ((_poly EQ NULL) OR (name EQ NULL));
	FATAL_ERROR_IF (_poly -> get_constraint NE get_STHGP_constraint);

	poly = (const struct STHGP *) _poly;

	s = match_mixed_case (name, sec_names);
	if (s NE NULL) {
		cp = get_STHGP_subtour (poly, s);
		return (cp);
	}

	s = match_mixed_case (name, nonneg_names);
	if (s NE NULL) {
		cp = get_STHGP_nonneg_constraint (poly, s);
		return (cp);
	}

	return (NULL);
}

/*
 * Parse the subtour size, and return either the correct subtour
 * inequality (for vertices 0...{k-1}, or NULL if the string is invalid.
 */

	static
	struct Constraint *
get_STHGP_subtour (

const struct STHGP *	poly,		/* IN: polytope to get subtour for */
const char *		s		/* IN: decimal string for K */
)
{
int			j, k, nedges, nz, count, v;
int **			edge_list;
int *			rmatind;
double *		rmatval;
int *			ip1;
int *			ip2;
struct Constraint *	cp;

	FATAL_ERROR_IF ((poly EQ NULL) OR (s EQ NULL));

	s = get_int (s, &k);		/* Scan and convert decimal number. */

	/* Check for syntax errors. */
	if (s EQ NULL) return (NULL);
	if (*s NE '\0') return (NULL);

	if ((k < 2) OR (k >= poly -> nverts)) {
		fprintf (stderr,
			 "Subtour size %d is out of range [2,%d]\n",
			 k, poly -> nverts - 1);
		return (NULL);
	}

	nedges		= poly -> nedges;
	edge_list	= poly -> edge_list;

	rmatind	= NEWA (nedges, int);
	rmatval	= NEWA (nedges, double);

	nz = 0;
	for (j = 0; j < nedges; j++) {
		count = 0;
		ip1 = edge_list [j];
		ip2 = edge_list [j + 1];
		while (ip1 < ip2) {
			v = *ip1++;
			if (v < k) {
				++count;
			}
		}
		if (count > 1) {
			rmatind [nz] = j;
			rmatval [nz] = count - 1;
			++nz;
		}
	}
	cp = make_constraint (nz, rmatind, rmatval, 'L', k - 1);

	free (rmatval);
	free (rmatind);

	return (cp);
}

/*
 * Parse the subtour size, and return either the correct non-negativity
 * inequality (for vertices 0...{k-1}, or NULL if the string is invalid.
 */

	static
	struct Constraint *
get_STHGP_nonneg_constraint (

const struct STHGP *	poly,		/* IN: polytope to get subtour for */
const char *		s		/* IN: decimal string for K */
)
{
int			j, k;
int			nedges;
int **			edge_list;
int *			ip1;
int *			ip2;
struct Constraint *	cp;
int			rmatind [1];
double			rmatval [1];

	FATAL_ERROR_IF ((poly EQ NULL) OR (s EQ NULL));

	s = get_int (s, &k);		/* Scan and convert decimal number. */

	/* Check for syntax errors. */
	if (s EQ NULL) return (NULL);
	if (*s NE '\0') return (NULL);

	if ((k < 2) OR (k > poly -> nverts)) {
		fprintf (stderr, "Edge size %d is out of range [2,%d]\n",
			 k, poly -> nverts);
		return (NULL);
	}

	nedges		= poly -> nedges;
	edge_list	= poly -> edge_list;

	for (j = 0; ; j++) {
		FATAL_ERROR_IF (j >= nedges);
		ip1 = edge_list [j];
		ip2 = edge_list [j + 1];
		if (ip2 - ip1 EQ k) {
			/* Found first k-vertex edge! */
			break;
		}
	}

	rmatind [0] = j;
	rmatval [0] = 1.0;

	cp = make_constraint (1, rmatind, rmatval, 'G', 0.0);

	return (cp);
}

/*
 * Return the edge_list data structure.
 */

	int **
get_STHGP_edge_list (

const struct Polytope *	_poly		/* IN: polytope to get edge_list of */
)
{
const struct STHGP *	poly;

	FATAL_ERROR_IF (_poly EQ NULL);
	FATAL_ERROR_IF (_poly -> free NE free_STHGP);

	poly = (const struct STHGP *) _poly;

	return (poly -> edge_list);
}

/*
 * Free up the given STHGP polytope.
 */

	static
	void
free_STHGP (

struct Polytope *	poly	/* IN: Polytope to free */
)
{
struct STHGP *		p;
int **			edge_list;
int *			ip;

	if (poly NE NULL) {
		FATAL_ERROR_IF (poly -> free NE free_STHGP);

		p = (struct STHGP *) poly;

		edge_list = p -> edge_list;
		if (edge_list NE NULL) {
			ip = edge_list [0];
			if (ip NE NULL) {
				free (ip);
			}
			free (edge_list);
		}
		p -> edge_list = NULL;

		clear_polytope (&(p -> base));
		free (p);
	}
}
