/***********************************************************************

	$Id: stirling2.c,v 1.1 2025/01/26 02:31:44 warme Exp $

	File:	stirling2.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

************************************************************************

	Stirling numbers of the second kind.

************************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "stirling2.h"

#include "fatal.h"
#include <gmp.h>
#include "gmpmisc.h"
#include "logic.h"
#include "new.h"
#include <stdlib.h>
#include <string.h>

/*
 * Local Types
 */

/*
 * An object to hold Stirling numbers of the second kind.
 */

struct Stirling2 {
	int		n_alloc;	/* Cardinality of largest set	*/
					/* this buffer has been		*/
					/* allocated to hold		*/
	int		n_fill;		/* Cardinality of largest set	*/
					/* currently computed		*/
	mpz_ptr *	rows;		/* Array of rows, one per set	*/
					/* cardinality			*/
};

/*
 * Local Functions
 */

static void			stirling2_fill (struct Stirling2 *	stp,
						int			n);
static struct Stirling2 *	stirling2_get (int n);

/*
 * Local Data
 */

static struct Stirling2 *	st2_buf;

/*
 * Return "n subset k", the Stirling number of the second kind.
 */

	void
stirling2 (

mpz_ptr			res,		/* IN/OUT: Result */
int			n,		/* IN: Set cardinality */
int			k		/* IN: Number of non-empty subsets */
)
{
struct Stirling2 *	stp;
mpz_ptr			row;

	FATAL_ERROR_IF (res EQ NULL);
	FATAL_ERROR_IF (n < 0);
	FATAL_ERROR_IF (k < 0);
	FATAL_ERROR_IF (k > n);

	stp = stirling2_get (n);

	row	= stp -> rows [n];

	mpz_set (res, &row [k]);
}

/*
 * Shut down stirling2 module, freeing all statically-held memory.
 */

	void
stirling2_shutdown (void)

{
int			i, n_alloc, n_fill;
struct Stirling2 *	stp;
mpz_ptr *		rows;
mpz_ptr			row;

	stp = st2_buf;
	if (stp NE NULL) {
		st2_buf	= NULL;

		n_alloc	= stp -> n_alloc;
		n_fill	= stp -> n_fill;
		rows	= stp -> rows;
		FATAL_ERROR_IF (n_alloc < 0);

		for (i = 0; i <= n_fill; i++) {
			row = rows [i];
			rows [i] = NULL;

			FATAL_ERROR_IF (row EQ NULL);

			free_Z_vector (row, i + 1);
		}
		free (rows);

		memset (stp, 0, sizeof (*stp));
		free (stp);
	}
}

/*
 * Return the Stirling2 buffer, assuring that all rows up to cardinality
 * n have been computed.
 */

	static
	struct Stirling2 *
stirling2_get (

int		n		/* IN: Cardinality required */
)
{
mpz_ptr *		rows;
struct Stirling2 *	stp;

	FATAL_ERROR_IF (n < 0);

	stp = st2_buf;
	if (stp EQ NULL) {
		rows = NEWA (n + 1, mpz_ptr);
		memset (rows, 0, (n + 1) * sizeof (rows [0]));

		stp = NEW (struct Stirling2);
		memset (stp, 0, sizeof (*stp));

		stp -> n_alloc	= n;
		stp -> n_fill	= -1;
		stp -> rows	= rows;

		st2_buf = stp;
	}

	/* Assure rows up to n have been filled. */
	stirling2_fill (stp, n);

	return (stp);
}

/*
 * Assure that all Stirling numbers have been calculated for all sets
 * up to size n.
 */

	static
	void
stirling2_fill (

struct Stirling2 *	stp,		/* IN/OUT: Stirling2 buffer to fill */
int			n		/* IN: Max cardinality set to fill */
)
{
int		i, k, n_alloc, n_fill;
mpz_ptr		row, prev_row;
mpz_ptr *	rows;
mpz_ptr *	rows2;
mpz_t		TEMP;

	FATAL_ERROR_IF (stp EQ NULL);
	FATAL_ERROR_IF (n < 0);

	n_alloc	= stp -> n_alloc;
	n_fill	= stp -> n_fill;
	rows	= stp -> rows;

	FATAL_ERROR_IF (n_fill > n_alloc);

	if (n > n_alloc) {
		/* Must re-allocate the buffer. */
		rows2 = NEWA (n + 1, mpz_ptr);
		memcpy (rows2, rows, (n_fill + 1) * sizeof (rows2 [0]));
		memset (&rows2 [n_fill + 1],
			0,
			(n - n_fill) * sizeof (rows2 [0]));
		stp -> n_alloc	= n;
		stp -> rows	= rows2;
		memset (rows, 0, (n_fill + 1) * sizeof (rows [0]));
		free (rows);
		rows = rows2;
	}

	if (n > n_fill) {
		mpz_init (TEMP);

		i = n_fill;
		while (i < n) {
			++i;
			row = new_Z_vector (i + 1);
			mpz_set_ui (&row [i], 1);
			if (i > 0) {
				/* Initialize row i > 0. */
				prev_row = rows [i - 1];
				mpz_set_ui (&row [0], 0);
				for (k = 1; k < i; k++) {
					mpz_mul_ui (TEMP, &prev_row [k], k);
					mpz_add (&row [k], TEMP, &prev_row [k - 1]);
				}
			}
			rows [i] = row;
		}
		stp -> n_fill = n;

		mpz_clear (TEMP);
	}
}
