/***********************************************************************

	File:	tsp_polytope.c
	Rev:	a-2
	Date:	01/25/2025

	Copyright (c) 1997, 2025 by David M. Warme

************************************************************************

	A wrapper for polytope TSP(n).

************************************************************************

	Modification Log:

	a-1:	09/07/2015	warme
		: Created.
	a-2:	01/25/2025	warme
		: Cleanup for initial release.

************************************************************************/

#include "tsp_polytope.h"

#include "bitmaskmacros.h"
#include "fatal.h"
#include "logic.h"
#include "new.h"
#include "polytope.h"
#include "s3itypes.h"
#include "scanutils.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Global Routines
 */

struct Polytope *	make_TSP_polytope (int n);

extern char *		me;

/*
 * The actual implementation of the TSP(n) Polytope.
 * This is a "derived class" from Polytope.
 */

struct TSP {
	struct Polytope	base;		/* The base class instance */
	int		nverts;		/* Number of vertices */
	int		nedges;		/* Number of edges */
	int *		edge_v1;	/* First verex of each edge */
	int *		edge_v2;	/* Second vertex of each edge */
};

struct gen {			/* For generating a set of tours */
	int		nverts;	/* Number of vertices */
	int		nedges;	/* Number of edges */
	int		nmasks;	/* Number of edge masks */
	int *		base;	/* Start of tour buffer */
	int *		endp;	/* End of tour buffer */
	struct tour *	list;	/* List of tours */
	struct tour **	hookp;	/* Hook pointer for adding to list */
};

struct tour {			/* Linked list of tours */
	bitmap_t *	tour;	/* Incidence vector for this tour */
	struct tour *	next;	/* Next tour in list */
};

/*
 * Local Routines
 */

static const char **	create_edge_names (int		nverts,
					   int		nedges,
					   int *	edge_v1,
					   int *	edge_v2);
static struct IVecSet *	enumerate_tours (const struct Polytope * poly);
static int		find_edge (int n, int i, int j);
static void		free_TSP (struct Polytope * poly);
static int		generate_edges (int **	edge_v1,
					int **	edge_v2,
					int	nverts);
static struct Constraint *
			get_TSP_3_toothed_comb_constraint (
					const struct TSP *	poly,
					const char *		s);
static struct Constraint *
			get_TSP_constraint (const struct Polytope *	poly,
					    const char *		name);
static struct Constraint *
			get_TSP_nonneg_constraint (
					const struct TSP *	poly,
					const char *		s);
static struct Constraint *
			get_TSP_subtour (const struct TSP *	poly,
					 const char *		s);
static void		recurse (struct gen *	genp,
				 int *		p);

/*
 * Return an object representing the polytope TSP(n).
 */

	struct Polytope *
make_TSP_polytope (

int		nverts		/* IN: number of vertices in complete graph */
				/*     from which tours are taken */
)
{
int			j, k;
int			nedges, maxlen;
struct TSP *		poly;
int *			edge_v1;
int *			edge_v2;
const char **		edge_names;
char			buffer [32];

	if (nverts < 3) {
		fprintf (stderr, "TSP(n) requires n >= 3.\n");
		return (NULL);
	}

	poly = NEW (struct TSP);
	memset (poly, 0, sizeof (*poly));

	nedges = generate_edges (&edge_v1, &edge_v2, nverts);

	edge_names = create_edge_names (nverts, nedges, edge_v1, edge_v2);

	sprintf (buffer, "TSP(%d)", nverts);

	maxlen = 0;
	for (j = 0; j < nedges; j++) {
		k = strlen (edge_names [j]);
		if (k > maxlen) {
			maxlen = k;
		}
	}

	poly -> base.name	= strdup (buffer);
	poly ->	base.dim	= nedges;
	poly -> base.ename	= edge_names;
	poly -> base.maxlen	= maxlen;
	poly -> base.get_extreme_points		= enumerate_tours;
	poly -> base.get_constraint		= get_TSP_constraint;
	poly -> base.free			= free_TSP;

	poly -> nverts		= nverts;
	poly -> nedges		= nedges;
	poly -> edge_v1		= edge_v1;
	poly -> edge_v2		= edge_v2;

	return (&(poly -> base));
}

/*
 * Generate the edges of the complete graph.
 */

	static
	int
generate_edges (

int **		_edge_v1,	/* IN/OUT: First vertex of each edge */
int **		_edge_v2,	/* IN/OUT: Second vertex of each edge */
int		nverts		/* IN: Number of vertices */
)
{
int		i, j, k, nedges;
int *		edge_v1;
int *		edge_v2;

	nedges = nverts * (nverts - 1) / 2;

	edge_v1 = NEWA (nedges, int);
	edge_v2 = NEWA (nedges, int);

	k = 0;
	for (i = 0; i < nverts; i++) {
		for (j = i + 1; j < nverts; j++) {
			edge_v1 [k] = i;
			edge_v2 [k] = j;
			++k;
		}
	}
	FATAL_ERROR_IF (k NE nedges);

	*_edge_v1 = edge_v1;
	*_edge_v2 = edge_v2;

	return (nedges);
}

/*
 * Create names for each of the edges.
 */

	static
	const char **
create_edge_names (

int		nverts,		/* IN: number of vertices */
int		nedges,		/* IN: number of edges */
int *		edge_v1,	/* IN: first vertex for each edge */
int *		edge_v2		/* IN: second vertex for each edge */
)
{
int		j;
const char **	edge_names;
char *		s;
char		buffer [64];

	edge_names = NEWA (nedges, const char *);

	if (nverts <= 10) {
		/* We can use single digits for each vertex. */
		for (j = 0; j < nedges; j++) {
			s = &buffer [0];
			*s++ = 'x';
			*s++ = '0' + edge_v1 [j];
			*s++ = '0' + edge_v2 [j];
			*s = '\0';
			edge_names [j] = strdup (buffer);
		}
	}
	else {
		/* At least one vertex requires more than 1 digit. */
		for (j = 0; j < nedges; j++) {
			sprintf (buffer, "x%d_%d", edge_v1 [j], edge_v2 [j]);
			edge_names [j] = strdup (buffer);
		}
	}
	return (edge_names);
}

/*
 * Enumerate all tours.
 */

	static
	struct IVecSet *
enumerate_tours (

const struct Polytope *	_poly	/* IN: polytope to enumerate all tours of */
)
{
int			i;
int			nverts, nedges, nmasks, ntours;
const struct TSP *	poly;
int *			tourbuf;
bitmap_t *		array;
bitmap_t *		tp;
struct tour *		p;
struct IVecSet *	xpts;
struct gen		gen;

	FATAL_ERROR_IF (_poly EQ NULL);
	FATAL_ERROR_IF (_poly -> get_extreme_points NE enumerate_tours);

	poly = (const struct TSP *) _poly;

	nverts		= poly -> nverts;
	nedges		= poly -> nedges;

	nmasks = BMAP_ELTS (nedges);

	FATAL_ERROR_IF (poly -> base.dim NE nedges);

	tourbuf = NEWA (nverts, int);
	for (i = 0; i < nverts; i++) {
		tourbuf [i] = i;
	}

	gen.nverts	= nverts;
	gen.nedges	= nedges;
	gen.nmasks	= nmasks;
	gen.base	= &tourbuf [0];
	gen.endp	= &tourbuf [nverts];
	gen.list	= NULL;
	gen.hookp	= &gen.list;

	/* tourbuf[0] always contains vertex 0. */
	recurse (&gen, &tourbuf [1]);

	ntours = 0;
	for (p = gen.list; p NE NULL; p = p -> next) {
		++ntours;
	}

	array = NEWA (ntours * nmasks, bitmap_t);
	memset (array, 0, ntours * nmasks * sizeof (array [0]));

	tp = array;
	for (p = gen.list; p NE NULL; p = p -> next, tp += nmasks) {
		memcpy (tp, p -> tour, nmasks * sizeof (bitmap_t));
	}

	for (;;) {
		p = gen.list;
		if (p EQ NULL) break;
		gen.list = p -> next;
		p -> next = NULL;
		free (p -> tour);
		free (p);
	}
	free (tourbuf);

	xpts = NEW (struct IVecSet);

	xpts ->	nvecs	= ntours;
	xpts -> dim	= nedges;
	xpts -> nmasks	= nmasks;
	xpts -> array	= array;

	return (xpts);
}

/*
 * Recursive portion of tour generation.
 */

	static
	void
recurse (

struct gen *	genp,
int *		p
)
{
int		i, j, u, v, e;
int		nverts, nmasks;
int *		base;
int *		endp;
int *		q;
bitmap_t *	tbuf;
struct tour *	tp;

	base	= genp -> base;
	endp	= genp -> endp;

	if (p >= endp) {
		nverts	= genp -> nverts;

		if (base [1] > base [nverts - 1]) {
			/* Skip mirror image tours. */
			return;
		}

		nmasks = genp -> nmasks;

		tbuf = NEWA (nmasks, bitmap_t);
		memset (tbuf, 0, nmasks * sizeof (tbuf [0]));

		for (i = 1; i < nverts; i++) {
			u = genp -> base [i - 1];
			v = genp -> base [i];
			e = find_edge (nverts, u, v);
			SETBIT (tbuf, e);
		}
		u = genp -> base [0];
		v = genp -> base [nverts - 1];
		e = find_edge (nverts, u, v);
		SETBIT (tbuf, e);

		tp = NEW (struct tour);
		tp -> tour = tbuf;
		tp -> next = NULL;
		*(genp -> hookp) = tp;
		genp -> hookp = &(tp -> next);

		return;
	}

	for (q = p; q < endp; q++) {
		i = *p;
		j = *q;
		*p = j;
		*q = i;
		recurse (genp, p+1);
		*p = i;
		*q = j;
	}
}

/*
 * Given number of vertices N, and vertices (i,j), return the index of
 * edge e = (i,j).
 */

	static
	int
find_edge (

int		n,		/* IN: number of vertices */
int		i,		/* IN: first vertex */
int		j		/* IN: second vertex */
)
{
int		k;

	FATAL_ERROR_IF ((i < 0) OR (i >= n));
	FATAL_ERROR_IF ((j < 0) OR (j >= n));
	FATAL_ERROR_IF (i EQ j);

	if (i > j) {
		k = i;
		i = j;
		j = k;
	}

	k = (-i + (n + n - 3)) * i;
	k = (k >> 1) + j - 1;

	return (k);
}

/*
 * Decode the given name and return the corresponding inequality.
 */

	static
	struct Constraint *
get_TSP_constraint (

const struct Polytope *	_poly,		/* IN: polytope to get constraint of */
const char *		name		/* IN: name of constraint */
)
{
const struct TSP *	poly;
const char *		s;
struct Constraint *	cp;

	static const char *	sec_names [] = {
		"sec-", "sec", "subtour-", "subtour", NULL
	};
	static const char *	nonneg_names [] = {
		"nn-", "nn", "nonneg-", "nonneg", NULL
	};
	static const char *	comb3_names [] = {
		"3comb-", NULL
	};

	FATAL_ERROR_IF ((_poly EQ NULL) OR (name EQ NULL));
	FATAL_ERROR_IF (_poly -> get_constraint NE get_TSP_constraint);

	poly = (const struct TSP *) _poly;

	s = match_mixed_case (name, sec_names);
	if (s NE NULL) {
		cp = get_TSP_subtour (poly, s);
		return (cp);
	}

	s = match_mixed_case (name, nonneg_names);
	if (s NE NULL) {
		cp = get_TSP_nonneg_constraint (poly, s);
		return (cp);
	}

	s = match_mixed_case (name, comb3_names);
	if (s NE NULL) {
		cp = get_TSP_3_toothed_comb_constraint (poly, s);
		return (cp);
	}

	return (NULL);
}

/*
 * Parse the subtour size, and return either the correct subtour
 * inequality (for vertices 0...{k-1}, or NULL if the string is invalid.
 */

	static
	struct Constraint *
get_TSP_subtour (

const struct TSP *	poly,		/* IN: polytope to get subtour for */
const char *		s		/* IN: decimal string for K */
)
{
int			j, k, nedges, nz, u, v;
int *			edge_v1;
int *			edge_v2;
int *			rmatind;
double *		rmatval;
struct Constraint *	cp;

	FATAL_ERROR_IF ((poly EQ NULL) OR (s EQ NULL));

	s = get_int (s, &k);		/* Scan and convert decimal number. */

	/* Check for syntax errors. */
	if (s EQ NULL) return (NULL);
	if (*s NE '\0') return (NULL);

	if ((k < 2) OR (k > poly -> nverts - 2)) {
		fprintf (stderr,
			 "Subtour size %d is out of range [2,%d]\n",
			 k, poly -> nverts - 2);
		return (NULL);
	}

	nedges		= poly -> nedges;
	edge_v1		= poly -> edge_v1;
	edge_v2		= poly -> edge_v2;

	rmatind	= NEWA (nedges, int);
	rmatval	= NEWA (nedges, double);

	nz = 0;
	for (j = 0; j < nedges; j++) {
		u = edge_v1 [j];
		v = edge_v2 [j];
		if ((u < k) AND (v >= k)) {
			rmatind [nz] = j;
			rmatval [nz] = 1.0;
			++nz;
		}
	}
	cp = make_constraint (nz, rmatind, rmatval, 'G', 2);

	free (rmatval);
	free (rmatind);

	return (cp);
}

/*
 * Generate the non-negativity constraint.
 */

	static
	struct Constraint *
get_TSP_nonneg_constraint (

const struct TSP *	poly,		/* IN: polytope to get nonneg for */
const char *		s		/* IN: additional text of name */
)
{
struct Constraint *	cp;
int			rmatind [1];
double			rmatval [1];

	FATAL_ERROR_IF ((poly EQ NULL) OR (s EQ NULL));

	/* Must be nothing else left in the name. */
	if (s EQ NULL) return (NULL);
	if (*s NE '\0') return (NULL);

	/* The first edge will do nicely. */
	rmatind [0] = 0;
	rmatval [0] = 1.0;

	cp = make_constraint (1, rmatind, rmatval, 'G', 0.0);

	return (cp);
}

/*
 * Parse the name of the particular 3-toothed comb inequality that
 * is desired.  There are lots of numbers running around here.
 *
 * The comb has the following structure:
 *
 *		0-----1
 *		|
 *		2-----3		7
 *		|
 *		4-----5
 *		|
 *		6
 *
 * The string must have the following syntax:
 *
 *	n0-n1-n2-n3-n4-n5-n6
 *
 * Where n0 through n6 are decimal integers giving the sizes of
 * sets 0 through 6 in the diagram.  The n7 count is obtained
 * implicitly as n7 = n - n0 - n1 - n2 - n3 - n4 - n5 - n6;
 * The given parameters must all be >= 1, except for n6 >= 0.
 *
 * Returns the inequality, or NULL if the string is invalid.
 */

	static
	struct Constraint *
get_TSP_3_toothed_comb_constraint (

const struct TSP *	poly,		/* IN: polytope to get subtour for */
const char *		s		/* IN: decimal string for K */
)
{
int			i, j, nverts, nedges, u, v;
int			total, nz, end, count, rhs;
int *			edge_v1;
int *			edge_v2;
int *			vclass;
int *			rmatind;
double *		rmatval;
struct Constraint *	cp;
int			parms [8];

	FATAL_ERROR_IF ((poly EQ NULL) OR (s EQ NULL));

	nverts = poly -> nverts;
	nedges = poly -> nedges;

	for (i = 0; i < 6; i++) {
		s = get_int (s, &parms [i]);

		if (s EQ NULL) return (NULL);
		if (*s NE '-') return (NULL);
		++s;
	}
	s = get_int (s, &parms [6]);

	/* Check for syntax errors. */
	if (s EQ NULL) return (NULL);
	if (*s NE '\0') return (NULL);

	total = 0;
	for (i = 0; i < 6; i++) {
		if (parms [i] < 1) {
			fprintf (stderr,
				 "%s: 3comb parameter %d out of range.\n",
				 me, i + 1);
			return (NULL);
		}
		total += parms [i];
	}
	if (parms [6] < 0) {
		fprintf (stderr, "%s: 3comb parameter 7 out of range.\n", me);
		return (NULL);
	}
	total += parms [6];
	if (total > nverts) {
		fprintf (stderr,
			 "%s: total of 3comb parameters exceed"
			 " number of vertices.\n", me);
		return (NULL);
	}
	parms [7] = nverts - total;

	/* Partition the vertices into the 8 vertex classes. */
	vclass = NEWA (nverts, int);
	v = 0;
	for (i = 0; i < 8; i++) {
		end = v + parms [i];
		for (; v < end; v++) {
			vclass [v] = i;
		}
	}
	FATAL_ERROR_IF (v NE nverts);

	edge_v1		= poly -> edge_v1;
	edge_v2		= poly -> edge_v2;

	rmatind	= NEWA (nedges, int);
	rmatval	= NEWA (nedges, double);

	nz = 0;
	for (j = 0; j < nedges; j++) {
		u = vclass [edge_v1 [j]];
		v = vclass [edge_v2 [j]];
		count = 0;
		if (((u & 0x01) EQ 0) AND (u <= 6) AND
		    ((v & 0x01) EQ 0) AND (v <= 6)) {
			/* Edge j lies in handle {0,2,4,6}. */
			++count;
		}
		for (i = 0; i <= 4; i += 2) {
			if (((u & 0xFFFFFFFE) EQ i) AND
			    ((v & 0xFFFFFFFE) EQ i)) {
				/* Edge j lies in tooth i. */
				++count;
			}
		}
		if (count > 0) {
			rmatind [nz] = j;
			rmatval [nz] = count;
			++nz;
		}
	}

	rhs = (parms [0] + parms [2] + parms [4] + parms [6])
	      + (parms [0] + parms [1])
	      + (parms [2] + parms [3])
	      + (parms [4] + parms [5])
	      - 5;		/* 5 = (3*t+1)/2, where t=3 teeth. */
	cp = make_constraint (nz, rmatind, rmatval, 'L', rhs);

	free (rmatval);
	free (rmatind);
	free (vclass);

	return (cp);
}

/*
 * Free up the given TSP polytope.
 */

	static
	void
free_TSP (

struct Polytope *	_poly	/* IN: Polytope to free */
)
{
struct TSP *		poly;

	if (_poly NE NULL) {
		FATAL_ERROR_IF (_poly -> free NE free_TSP);

		poly = (struct TSP *) _poly;

		if (poly -> edge_v1 NE NULL) {
			free (poly -> edge_v1);
			poly -> edge_v1 = NULL;
		}

		if (poly -> edge_v2 NE NULL) {
			free (poly -> edge_v2);
			poly -> edge_v2 = NULL;
		}

		clear_polytope (&(poly -> base));
		free (poly);
	}
}
