/**********************************************************************

	$Id: gen_CD_poly.c,v 1.1 2025/01/26 02:31:51 warme Exp $

	File:	gen_CD_poly.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Generate an xfig file containing a line drawing of a
	polytope on which the Centroid Distance indicator performs
	poorly.

***********************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

***********************************************************************/

/*
 * This program generates an "xfig" file containing a line drawing of
 * the polytope that serves as a counter-example to the "centroid distance"
 * strength metric.
 *
 * This is a regular tetrahedron whose vertices have been replaced with
 * more complex "rounded off" polytopes.  There is a "small angle ring"
 * that isolates the main faces of the tetrahedron from the "tip"
 * polytopes.
 */

#include "fatal.h"
#include "logic.h"
#include <math.h>
#include "new.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

char *		me;

/*
 * Local Constants
 */

#define	CAMERA_X	-8.0
#define	CAMERA_Y	-6.0
#define	CAMERA_Z	3.0

#define	VIEWPORT_ALPHA	0.5

#define VIEWPORT_WORLD_HEIGHT	0.25
#define	VIEWPORT_WORLD_WIDTH	0.25

#define	VIEWPORT_DEV_HEIGHT	1200.0
#define	VIEWPORT_DEV_WIDTH	1200.0

#define	XFIG_RESOLUTION		1200.0

#define	IMAGE_CENTER_X		4.0
#define	IMAGE_CENTER_Y		3.0

#define	TRUNC_EPS		0.2

/*
 * Local Types
 */

struct Edge;
struct EList;
struct Facet;
struct Point2;
struct Point3;
struct Vertex;

enum Style {
	STYLE_NORMAL,		/* No fill */
	STYLE_ISOLATION,	/* Isolation facets */
	STYLE_LAYER_1,		/* Layer 1 facets */
	STYLE_LAYER_2,		/* Layer 2 facets */
	STYLE_END_CAP,		/* End Cap facets */
};

struct Point3 {			/* 3-dimensional point */
	double	d_x;
	double	d_y;
	double	d_z;
};

struct Point2 {			/* 2-dimensional point */
	double	d_x;
	double	d_y;
};

struct Vertex {
	struct Point3		d_point;	/* Coordinates of vertex */
	struct EList *		d_edges;	/* List of incident edges */
	struct Vertex *		d_next;		/* Next vertex in list */
	int			d_index;	/* Index of this vertex */
};

struct EList {
	struct Edge *		d_edge;		/* Incident edge */
	struct EList *		d_next;		/* Next incident edge */
	struct EList *		d_forw;		/* Link for all ELists */
};

struct Edge {			/* Linked list of 3-d edges (line segments) */
	struct Vertex *		d_v1;		/* First vertex / endpoint */
	struct Vertex *		d_v2;		/* Second vertex / endpoint */
	struct Facet *		d_f1;		/* First incident facet */
	struct Facet *		d_f2;		/* Second incident facet */
	struct Edge *		d_next;		/* Next Edge in list */
	int			d_index;	/* Index of this Edge */
};

struct Facet {			/* Linked list of 3-d polygons */
	int			d_nverts;	/* Number of vertices */
	int			d_index;	/* Index of this Facet */
	enum Style		d_style;	/* Style to display with */
	struct Vertex **	d_verts;	/* Array of vertices */
	struct Edge **		d_edges;	/* Incident edges */
	struct Point3		d_normal;	/* Unit normal vector */
	struct Facet *		d_next;		/* Next facet in list */
};

/*
 * Local Functions
 */

static void	cross_prod (struct Point3 *		dst,
			    const struct Point3 *	p1,
			    const struct Point3 *	p2);
static double	dot3 (const struct Point3 * p1, const struct Point3 * p2);
static void	emit_xfig_centroid (void);
static void	emit_xfig_header (void);
static void	emit_xfig_edge (const struct Edge * p);
static void	emit_xfig_facet (const struct Facet * fp);
static void	free_edges (void);
static void	free_ELists (void);
static void	free_facets (void);
static void	free_vertices (void);
static void	init_one_ttvert (struct Vertex **	verts,
				 const struct Point3 *	apex,
				 const struct Point3 *	v0,
				 const struct Point3 *	v1,
				 const struct Point3 *	v2);
static void	init_ttverts (void);
static void	init_tpoints (void);
static void	init_viewport (void);
static void	interp_seg (struct Point3 *		dst,
			    const struct Point3 *	p1,
			    const struct Point3 *	p2,
			    double			alpha);
static double	mag3 (const struct Point3 * src);
static struct Edge *
		make_edge (struct Vertex * v1, struct Vertex * v2);
static struct EList *
		make_elist (struct Edge * ep, struct EList * next);
static struct Facet *
		make_facet (int nverts, struct Vertex ** verts);
static void	make_truncation_polytope (const struct Point3 *	apex,
					  struct Vertex **	verts);
static struct Vertex *
		make_vertex (const struct Point3 * pnt);
static void	project_onto_viewport (struct Point2 *		dst,
				       const struct Point3 *	src);
static void	scaled_tetrahedron_section (
				struct Point3 *		result,
				const struct Point3 *	base,
				const struct Point3 *	apex,
				double			alpha,
				double			scale);
static void	unitize3 (struct Point3 *	dst,
			  const struct Point3 *	src);

/*
 * Local Variables
 */

static struct Point3	camera = {CAMERA_X, CAMERA_Y, CAMERA_Z};
static struct Point3	camera_dir;		/* Unit vector from origin */
						/* toward camera */
static double		camera_mag;		/* Magnitude of camera point */
static int		edge_index;		/* Index of next Edge */
static struct Edge *	edge_list;		/* List of all Edge objects */
static struct EList *	elist_list;		/* List of all EList objects */
static int		facet_index;		/* Index of next Facet */
static struct Facet *	facet_list;		/* List of all Facet objects */
static struct Point3	tpoints [4];		/* Tetrahedron vertex coords */
static struct Vertex *	ttverts [4] [3];	/* Truncated tetrahedron verts */
static struct Point3	vp_xdir;		/* Unit vector in viewport X direction */
static struct Point3	vp_ydir;		/* Unit vactor in viewport Y direction */
static int		vertex_index;		/* Index of next Vertex */
static struct Vertex *	vertex_list;		/* List of all Vertex objects */

/*
 * The main routine.
 */
	int
main (

int		argc,
char **		argv
)
{
int			i;

	me = argv [0];

	if (argc >= 4) {
		camera.d_x = atof (argv [1]);
		camera.d_y = atof (argv [2]);
		camera.d_z = atof (argv [3]);
	}

	emit_xfig_header ();

	camera_mag = mag3 (&camera);

	unitize3 (&camera_dir, &camera);

	init_viewport ();

	init_tpoints ();

	init_ttverts ();

	/* Make first tetrahedron Facet. */
	struct Vertex *		verts [6];
	verts [0] = ttverts [0] [0];
	verts [1] = ttverts [0] [2];
	verts [2] = ttverts [2] [0];
	verts [3] = ttverts [2] [2];
	verts [4] = ttverts [1] [1];
	verts [5] = ttverts [1] [0];
	make_facet (6, &verts [0]);

	/* Make 2nd tetrahedron Facet. */
	verts [0] = ttverts [0] [0];
	verts [1] = ttverts [1] [0];
	verts [2] = ttverts [1] [2];
	verts [3] = ttverts [3] [1];
	verts [4] = ttverts [3] [0];
	verts [5] = ttverts [0] [1];
	make_facet (6, &verts [0]);

	/* Make 3rd tetrahedron Facet. */
	verts [0] = ttverts [0] [1];
	verts [1] = ttverts [3] [0];
	verts [2] = ttverts [3] [2];
	verts [3] = ttverts [2] [1];
	verts [4] = ttverts [2] [0];
	verts [5] = ttverts [0] [2];
	make_facet (6, &verts [0]);

	/* Make 4th tetrahedron Facet. */
	verts [0] = ttverts [1] [1];
	verts [1] = ttverts [2] [2];
	verts [2] = ttverts [2] [1];
	verts [3] = ttverts [3] [2];
	verts [4] = ttverts [3] [1];
	verts [5] = ttverts [1] [2];
	make_facet (6, &verts [0]);

	/* Make the four truncation polytopes. */
	for (i = 0; i < 4; i++) {
		make_truncation_polytope (&tpoints [i], &ttverts [i] [0]);
	}

	struct Facet * fp;
	for (fp = facet_list; fp NE NULL; fp = fp -> d_next) {
		emit_xfig_facet (fp);
	}

	emit_xfig_centroid ();

	/* Muzzle compiler warning. */
	(void) emit_xfig_edge;

	free_vertices ();
	free_ELists ();
	free_edges ();
	free_facets ();

	exit (0);
}

/*
 * Emit the .fig file header.
 */

	static
	void
emit_xfig_header (void)

{
	printf ("#FIG 3.2  Produced by gen_CD_poly.c\n"
		"Landscape\n"
		"Center\n"
		"Inches\n"
		"Letter\n"
		"100.0\n"
		"Single\n"
		"-2\n"
		"1200 2\n");
}

/*
 * Emit an xfig line for the polygon associated with the given Facet.
 */

	static
	void
emit_xfig_facet (

const struct Facet *		fp
)
{
int		i, n, fill_color, area_fill;
struct Point3 *	p;
struct Vertex *	vp;
struct Point3	center, cdir;
struct Point2	pnt2;

	FATAL_ERROR_IF (fp EQ NULL);

	n = fp -> d_nverts;

#if 1
	/* Compute "center" of Facet polygon. */
	memset (&center, 0, sizeof (center));
	for (i = 0; i < n; i++) {
		p = &(fp -> d_verts [i] -> d_point);
		center.d_x += p -> d_x;
		center.d_y += p -> d_y;
		center.d_z += p -> d_z;
	}
	center.d_x /= n;
	center.d_y /= n;
	center.d_y /= n;

	/* Compute direction from center to camera. */
	cdir.d_x = camera.d_x - center.d_x;
	cdir.d_y = camera.d_y - center.d_y;
	cdir.d_z = camera.d_z - center.d_z;

	if (dot3 (&(fp -> d_normal), &cdir) <= 0.0) {
		/* This facet is oriented away from the camera,	*/
		/* and is therefore not visible.		*/
		return;
	}
#endif

	fill_color	= 7;	/* White */
	area_fill	= 20;	/* Full color intensity */

	switch (fp -> d_style) {
	case STYLE_NORMAL:
		fill_color = 7;		/* White */
		break;

	case STYLE_ISOLATION:
		fill_color = 30;	/* Light pink */
		break;

	case STYLE_LAYER_1:
		fill_color = 17;	/* Light cyan */
		break;

	case STYLE_LAYER_2:
		fill_color = 23;	/* Light magenta */
		break;

	case STYLE_END_CAP:
		fill_color = 2;		/* Green */
		break;

	default:
		FATAL_ERROR;
		break;
	}

	/* PolyLine attributes. */
	printf ("2 3 0 1 0 %d 50 -1 %d 0.000 1 1 -1 0 0 %d\n",
		fill_color, area_fill, n + 1);
	printf ("\t");
	for (i = 0; i < n; i++) {
		vp = fp -> d_verts [i];
		project_onto_viewport (&pnt2, &(vp -> d_point));
		printf (" %.17g %.17g",
			floor (pnt2.d_x + 0.5),
			floor (pnt2.d_y + 0.5));
	}
	vp = fp -> d_verts [0];
	project_onto_viewport (&pnt2, &(vp -> d_point));
	printf (" %.17g %.17g\n",
		floor (pnt2.d_x + 0.5),
		floor (pnt2.d_y + 0.5));
}

/*
 * Emit a black, filled circle to represent the centroid.
 */

	static
	void
emit_xfig_centroid (void)

{
int		depth;
double		r;
struct Point3	Origin;
struct Point2	C;

	depth = 49;

	memset (&Origin, 0, sizeof (Origin));

	project_onto_viewport (&C, &Origin);

	r = 60.0;
	C.d_x = floor (C.d_x + 0.5);
	C.d_y = floor (C.d_y + 0.5);

	printf ("1 3 0 1 0 0 %d -1 20 0.000 1 0.0000"
		" %.17g %.17g %.17g %.17g %.17g %.17g %.17g %.17g\n",
		depth, C.d_x, C.d_y, r, r, C.d_x, C.d_y, C.d_x, C.d_y + r);

	/* Now label the centroid. */
	C.d_x += 100.0;
	C.d_y += 200.0;
	printf ("4 0 0 %d -1 1 12.0 0.000 0 75.0 75.0 %.17g %.17g %s\001\n",
		depth, C.d_x, C.d_y, "C");
}

/*
 * Emit an xfig line for the given "world coordinates" line.
 */

	static
	void
emit_xfig_edge (

const struct Edge *		ep
)
{
struct Point2		p1, p2;

	project_onto_viewport (&p1, &(ep -> d_v1 -> d_point));
	project_onto_viewport (&p2, &(ep -> d_v2 -> d_point));

	/* PolyLine attributues. */
	printf ("2 1 0 1 0 7 50 -1 -1 0.000 0 0 -1 0 0 2\n");
	printf ("\t %.17g %.17g %.17g %.17g\n",
		floor (p1.d_x + 0.5),
		floor (p1.d_y + 0.5),
		floor (p2.d_x + 0.5),
		floor (p2.d_y + 0.5));
}

/*
 * Init tpoints[] -- the coordinates of the vertices of the unit tetrahedron.
 */

	static
	void
init_tpoints (void)

{
struct Point3 *		p;
double			sqrt_2, sqrt_3;

	sqrt_2 = sqrt (2.0);
	sqrt_3 = sqrt (3.0);

	/* Initialize each vertex (as a unit vector). */
	p = &tpoints [0];

	p -> d_x	= sqrt_2 * 2.0 / 3.0;
	p -> d_y	= 0.0;
	p -> d_z	= - 1.0 / 3.0;
	++p;

	p -> d_x	= - sqrt_2 / 3.0;
	p -> d_y	= sqrt_2 / sqrt_3;
	p -> d_z	= -1.0 / 3.0;
	++p;

	p -> d_x	= - sqrt_2 / 3.0;
	p -> d_y	= - sqrt_2 / sqrt_3;
	p -> d_z	= - 1.0 / 3.0;
	++p;

	p -> d_x	= 0.0;
	p -> d_y	= 0.0;
	p -> d_z	= sqrt_2 * 2.0 / 3.0;
}

/*
 * Initialize ttverts[][], the vertices of the tetrahedron "truncation"
 * facets.
 */

	static
	void
init_ttverts (void)

{
const struct Point3	*	v0;
const struct Point3	*	v1;
const struct Point3	*	v2;
const struct Point3	*	v3;

	v0 = &tpoints [0];
	v1 = &tpoints [1];
	v2 = &tpoints [2];
	v3 = &tpoints [3];

	init_one_ttvert (&ttverts [0] [0], v0, v1, v3, v2);

	init_one_ttvert (&ttverts [1] [0], v1, v0, v2, v3);

	init_one_ttvert (&ttverts [2] [0], v2, v0, v3, v1);

	init_one_ttvert (&ttverts [3] [0], v3, v0, v1, v2);
}

/*
 * Construct the points of the Facet polygon that truncates the given
 * apex point having neighboring points v0, v1, and v2, which are
 * given in counter-clockwise order with respect to the outward-facing
 * normal vector of said Facet.
 */

	static
	void
init_one_ttvert (

struct Vertex **	verts,	/* IN/OUT: Three Vertices */
const struct Point3 *	apex,	/* IN: Apex point being truncated */
const struct Point3 *	v0,	/* IN: Neighboring vertex point v0 */
const struct Point3 *	v1,	/* IN: Neighboring vertex point v1 */
const struct Point3 *	v2	/* IN: Neighboring vertex point v2 */
)
{
struct Point3	pnt0, pnt1, pnt2;

	interp_seg (&pnt0, apex, v0, TRUNC_EPS);
	interp_seg (&pnt1, apex, v1, TRUNC_EPS);
	interp_seg (&pnt2, apex, v2, TRUNC_EPS);

	verts [0] = make_vertex (&pnt0);
	verts [1] = make_vertex (&pnt1);
	verts [2] = make_vertex (&pnt2);
}

/*
 * Interpolate between two points: dst = (1 - alpha) * p1 + alpha * p2.
 */

	static
	void
interp_seg (

struct Point3 *		dst,
const struct Point3 *	p1,
const struct Point3 *	p2,
double			alpha
)
{
double		beta;

	beta = 1.0 - alpha;

	dst -> d_x = (beta * p1 -> d_x) + (alpha * p2 -> d_x);
	dst -> d_y = (beta * p1 -> d_y) + (alpha * p2 -> d_y);
	dst -> d_z = (beta * p1 -> d_z) + (alpha * p2 -> d_z);
}

/*
 * Make the "truncation polytope" for the given apex point, attached to
 * the rest of the polytope by the 3 given vertices.
 */

	static
	void
make_truncation_polytope (

const struct Point3 *	apex,	/* IN: Apex being truncated */
struct Vertex **	verts	/* IN: 3 vertices at which we attach to	*/
				/*     the rest of the polytope (in	*/
				/*     counter-clockwise order around	*/
				/*     an outward-facing normal vector)	*/
)
{
#if 1
int			i;
struct Point3 *		p1;
struct Point3 *		p2;
struct Vertex *		v0;
struct Vertex *		v1;
struct Vertex *		v2;
struct Vertex *		vp1;
struct Vertex *		vp2;
struct Vertex *		vp3;
struct Vertex *		vp4;
struct Facet *		fp;
struct Vertex *		iverts [3];
struct Vertex *		jverts [3];
struct Vertex *		fverts [4];
struct Point3		base, tr1 [3], tr2 [3], tr3 [3], tr4 [3];

	v0 = verts [0];
	v1 = verts [1];
	v2 = verts [2];

	/* Make first prototype triangle tr1. */
	tr1 [0]	= v0 -> d_point;
	tr1 [1] = v1 -> d_point;
	tr1 [2] = v2 -> d_point;

	memset (&base, 0, sizeof (base));
	for (i = 0; i < 3; i++) {
		base.d_x += tr1 [i].d_x;
		base.d_y += tr1 [i].d_y;
		base.d_z += tr1 [i].d_z;
	}
	base.d_x /= 3.0;
	base.d_y /= 3.0;
	base.d_z /= 3.0;

	/* Make second prototype triangle.  It's vertices are the	*/
	/* midpoints of the edges of tr1.				*/
	p1 = &tr1 [2];
	for (i = 0; i < 3; i++) {
		p2 = &tr1 [i];
		tr2 [i].d_x = 0.5 * (p1 -> d_x + p2 -> d_x);
		tr2 [i].d_y = 0.5 * (p1 -> d_y + p2 -> d_y);
		tr2 [i].d_z = 0.5 * (p1 -> d_z + p2 -> d_z);
		p1 = p2;
	}

	/* Make "isolation" Facets.  These are quadrilaterals having a	*/
	/* very small angle with respect to their adjacent tetrahedron	*/
	/* Facet.							*/

	scaled_tetrahedron_section (&tr3 [0], &tr1 [0], apex, 0.9, 0.95);
	for (i = 0; i < 3; i++) {
		iverts [i] = make_vertex (&tr3 [i]);
	}
	vp1 = verts [2];
	vp2 = iverts [2];
	for (i = 0; i < 3; i++) {
		vp3 = verts [i];
		vp4 = iverts [i];
		fverts [0] = vp1;
		fverts [1] = vp3;
		fverts [2] = vp4;
		fverts [3] = vp2;

		fp = make_facet (4, &fverts [0]);
		fp -> d_style = STYLE_ISOLATION;

		vp1 = vp3;
		vp2 = vp4;
	}

	/* Make next layer of triangular Facets. */

	scaled_tetrahedron_section (&tr4 [0], &tr2 [0], apex, 0.6, 0.9);
	for (i = 0; i < 3; i++) {
		jverts [i] = make_vertex (&tr4 [i]);
	}

	vp1 = iverts [2];
	vp3 = jverts [2];
	for (i = 0; i < 3; i++) {
		vp2 = iverts [i];
		vp4 = jverts [i];
		fverts [0] = vp1;
		fverts [1] = vp2;
		fverts [2] = vp4;

		fp = make_facet (3, &fverts [0]);
		fp -> d_style = STYLE_LAYER_1;

		fverts [0] = vp3;
		fverts [1] = vp1;
		fverts [2] = vp4;

		fp = make_facet (3, &fverts [0]);
		fp -> d_style = STYLE_LAYER_2;

		vp1 = vp2;
		vp3 = vp4;
	}

	/* FIXME -- just cap the end for now. */
	fp = make_facet (3, &jverts [0]);
	fp -> d_style = STYLE_END_CAP;

#else
	make_facet (3, &verts [0]);
#endif
}

/*
 * We are given a tetrahedron with 3 "base" points and one "apex" point.
 * Compute the triangle that is the "cross section" of this tetrahedron
 * at fractional position of "alpha" (where alpha=0 at the apex, and
 * alpha=1 at the base).  Then scale this triangle by the given factor,
 * keeping the same center and planar alignment.
 */

	static
	void
scaled_tetrahedron_section (

struct Point3 *		result,		/* IN/OUT: resulting triangle points */
const struct Point3 *	base,		/* IN: Points of the base */
const struct Point3 *	apex,		/* IN: The apex point */
double			alpha,		/* IN: Frac pos of cross section */
double			scale		/* IN: Final scale factor */
)
{
int			i;
double			beta;
struct Point3		C, T [3];

	/* Compute cross section triangle. */
	beta = 1.0 - alpha;
	for (i = 0; i < 3; i++) {
		T [i].d_x = alpha * base [i].d_x + beta * apex -> d_x;
		T [i].d_y = alpha * base [i].d_y + beta * apex -> d_y;
		T [i].d_z = alpha * base [i].d_z + beta * apex -> d_z;
	}

	/* Compute center of this triangle. */
	memset (&C, 0, sizeof (C));
	for (i = 0; i < 3; i++) {
		C.d_x += T [i].d_x;
		C.d_y += T [i].d_y;
		C.d_z += T [i].d_z;
	}
	C.d_x /= 3.0;
	C.d_y /= 3.0;
	C.d_z /= 3.0;

	/* Scale triangle to get final result. */
	beta = 1.0 - scale;
	for (i = 0; i < 3; i++) {
		result [i].d_x = scale * T [i].d_x + beta * C.d_x;
		result [i].d_y = scale * T [i].d_y + beta * C.d_y;
		result [i].d_z = scale * T [i].d_z + beta * C.d_z;
	}
}

/*
 * Make a Facet for the given sequence of Vertex objects.
 * The vertices are required to be listed in counter-clockwise order
 * around the Facet, and the "right hand rule" gives a normal vector
 * pointing *outward* from the polytope.
 */

	static
	struct Facet *
make_facet (

int			nverts,		/* IN: Number of Vertices */
struct Vertex **	verts		/* IN: Array of Vertex ptrs */
)
{
int			i, j, k;
struct Vertex **	varray;
struct Vertex *		v1;
struct Vertex *		v2;
struct Vertex *		v3;
struct Edge **		earray;
struct Facet *		fp;
struct Edge *		ep;
double			mag;
struct Point3		pnt1, pnt2, pnt3, normal;

	FATAL_ERROR_IF (nverts < 3);
	FATAL_ERROR_IF (verts EQ NULL);

	for (i = 0; i < nverts; i++) {
		FATAL_ERROR_IF (verts [i] EQ NULL);
	}

	fp = NEW (struct Facet);
	memset (fp, 0, sizeof (*fp));

	varray	= NEWA (nverts, struct Vertex *);
	memcpy (varray, verts, nverts * sizeof (varray [0]));

	earray	= NEWA (nverts, struct Edge *);
	memset (earray, 0, nverts * sizeof (earray [0]));

	v1 = verts [nverts - 1];
	for (i = 0; i < nverts; v1 = v2, i++) {
		v2 = verts [i];
		ep = make_edge (v1, v2);

		if (ep -> d_v1 EQ v1) {
			FATAL_ERROR_IF (ep -> d_v2 NE v2);
			FATAL_ERROR_IF (ep -> d_f1 NE NULL);
			ep -> d_f1 = fp;
		}
		else if (ep -> d_v1 EQ v2) {
			FATAL_ERROR_IF (ep -> d_v2 NE v1);
			FATAL_ERROR_IF (ep -> d_f2 NE NULL);
			ep -> d_f2 = fp;
		}
		else {
			FATAL_ERROR;
		}

		earray [i] = ep;
	}

	/* Verify that all of the points are co-planar by computing the	*/
	/* cross-product unit vectors on all triples of consecutive	*/
	/* vertices.  These should all be the same unit normal vector.	*/
	memset (&normal, 0, sizeof (normal));
	k = nverts - 2;
	j = nverts - 1;
	for (i = 0; i < nverts; k = j, j = i, i++) {
		v1 = verts [k];
		v2 = verts [j];
		v3 = verts [i];

		pnt1.d_x = v1 -> d_point.d_x - v2 -> d_point.d_x;
		pnt1.d_y = v1 -> d_point.d_y - v2 -> d_point.d_y;
		pnt1.d_z = v1 -> d_point.d_z - v2 -> d_point.d_z;

		pnt2.d_x = v3 -> d_point.d_x - v2 -> d_point.d_x;
		pnt2.d_y = v3 -> d_point.d_y - v2 -> d_point.d_y;
		pnt2.d_z = v3 -> d_point.d_z - v2 -> d_point.d_z;

		cross_prod (&pnt3, &pnt2, &pnt1);

		unitize3 (&pnt3, &pnt3);
		if (i EQ 0) {
			normal = pnt3;
		}
		else {
			pnt1.d_x = pnt3.d_x - normal.d_x;
			pnt1.d_y = pnt3.d_y - normal.d_y;
			pnt1.d_z = pnt3.d_z - normal.d_z;
			mag = mag3 (&pnt1);
			FATAL_ERROR_IF (mag >= 1.0E-14);
		}
	}

	fp -> d_nverts		= nverts;
	fp -> d_index		= facet_index++;
	fp -> d_style		= STYLE_NORMAL;
	fp -> d_verts		= varray;
	fp -> d_edges		= earray;
	fp -> d_normal		= normal;
	fp -> d_next		= facet_list;

	facet_list = fp;

	return (fp);
}

/*
 * Make a new Vertex having the given 3-d coordinates.
 */

	static
	struct Vertex *
make_vertex (

const struct Point3 *	pnt	/* IN: Coordinates of new vertex */
)
{
struct Vertex *		vp;

	vp = NEW (struct Vertex);

	vp -> d_point	= *pnt;
	vp -> d_edges	= NULL;
	vp -> d_next	= vertex_list;
	vp -> d_index	= vertex_index++;

	vertex_list = vp;

	return (vp);
}

/*
 * Free up all of the Facet objects.
 */

	static
	void
free_facets (void)

{
struct Facet *		fp;

	for (;;) {
		fp = facet_list;
		if (fp EQ NULL) break;

		facet_list = fp -> d_next;

		free (fp -> d_verts);
		free (fp -> d_edges);

		memset (fp, 0, sizeof (*fp));
		free (fp);
	}
}

/*
 * Free up all of the Edge objects.
 */

	static
	void
free_edges (void)

{
struct Edge *		ep;

	for (;;) {
		ep = edge_list;
		if (ep EQ NULL) break;

		edge_list = ep -> d_next;

		memset (ep, 0, sizeof (*ep));
		free (ep);
	}
}

/*
 * Free up all of the Vertex objects.
 */

	static
	void
free_vertices (void)

{
struct Vertex *		p1;

	for (;;) {
		p1 = vertex_list;
		if (p1 EQ NULL) break;

		vertex_list	= p1 -> d_next;

		memset (p1, 0, sizeof (*p1));
		free (p1);
	}
}

/*
 * Free up all of the EList objects.
 */

	static
	void
free_ELists (void)

{
struct EList *		p1;
struct EList *		p2;

	p1 = elist_list;
	elist_list = NULL;

	while (p1 NE NULL) {
		p2 = p1 -> d_forw;

		memset (p1, 0, sizeof (*p1));
		free (p1);

		p1 = p2;
	}
}

/*
 * Return the unique Edge for the two given vertices, creating this
 * Edge if it does not yet exist.
 */

	static
	struct Edge *
make_edge (

struct Vertex *		v1,	/* IN: First Vertex */
struct Vertex *		v2	/* IN: Second Vertex */
)
{
struct Edge *		p;

	FATAL_ERROR_IF (v1 EQ NULL);
	FATAL_ERROR_IF (v2 EQ NULL);
	FATAL_ERROR_IF (v1 EQ v2);

	for (p = edge_list; p NE NULL; p = p -> d_next) {
		if ((p -> d_v1 EQ v1) AND (p -> d_v2 EQ v2)) {
			return (p);
		}
		if ((p -> d_v1 EQ v2) AND (p -> d_v2 EQ v1)) {
			return (p);
		}
	}

	/* No such Edge exists yet.  Create it now. */

	p = NEW (struct Edge);
	p -> d_v1	= v1;
	p -> d_v2	= v2;
	p -> d_f1	= NULL;
	p -> d_f2	= NULL;
	p -> d_next	= edge_list;
	p -> d_index	= edge_index++;

	edge_list = p;

	v1 -> d_edges = make_elist (p, v1 -> d_edges);
	v2 -> d_edges = make_elist (p, v2 -> d_edges);

	return (p);
}

/*
 * Make an EList object for given Edge and "rest of list."
 */

	static
	struct EList *
make_elist (

struct Edge *		ep,	/* IN: Edge to add to list */
struct EList *		next	/* IN: Rest of linked list of EList objects */
)
{
struct EList *		elp;

	FATAL_ERROR_IF (ep EQ NULL);

	elp = NEW (struct EList);

	elp -> d_edge	= ep;
	elp -> d_next	= next;
	elp -> d_forw	= elist_list;

	elist_list = elp;

	return (elp);
}

/*
 * Initialize variables used to project into the viewport rectangle.
 */

	static
	void
init_viewport (void)

{
struct Point3	up;

	up.d_x = 0.0;
	up.d_y = 0.0;
	up.d_z = 1.0;

	cross_prod (&vp_xdir, &up, &camera_dir);
	unitize3 (&vp_xdir, &vp_xdir);

	cross_prod (&vp_ydir, &camera_dir, &vp_xdir);
	unitize3 (&vp_ydir, &vp_ydir);
}

/*
 * Project the given 3-d "world coordinates" point to the corresponding
 * 2-d point within the viewport (in viewport device-specific coordinates).
 */

	static
	void
project_onto_viewport (

struct Point2 *		dst,
const struct Point3 *	src
)
{
double		num, den, alpha, shift, scale_x, scale_y;
struct Point3	proj;
struct Point2	vp_pnt;

	num = VIEWPORT_ALPHA * camera_mag;
	den = dot3 (&camera_dir, src) - camera_mag;

	FATAL_ERROR_IF (den EQ 0.0);

	alpha = num / den;

	proj.d_x = (1.0 - alpha) * camera.d_x + alpha * src -> d_x;
	proj.d_y = (1.0 - alpha) * camera.d_y + alpha * src -> d_y;
	proj.d_z = (1.0 - alpha) * camera.d_z + alpha * src -> d_z;

	/* Translate viewport plane so that its origin corresponds to	*/
	/* the word coordinates origin.					*/
	shift = (1.0 - alpha) * camera_mag;
	proj.d_x -= shift * camera_dir.d_x;
	proj.d_y -= shift * camera_dir.d_y;
	proj.d_z -= shift * camera_dir.d_z;

	scale_x = VIEWPORT_DEV_WIDTH  / VIEWPORT_WORLD_WIDTH;
	scale_y = VIEWPORT_DEV_HEIGHT / VIEWPORT_WORLD_HEIGHT;

	vp_pnt.d_x = scale_x * dot3 (&proj, &vp_xdir);
	vp_pnt.d_y = scale_y * dot3 (&proj, &vp_ydir);

	vp_pnt.d_x += IMAGE_CENTER_X * XFIG_RESOLUTION;
	vp_pnt.d_y += IMAGE_CENTER_Y * XFIG_RESOLUTION;

	*dst = vp_pnt;
}

/*
 * Compute the cross product:	dst = p1 x p2.
 */

	static
	void
cross_prod (

struct Point3 *		dst,
const struct Point3 *	p1,
const struct Point3 *	p2
)
{
	dst -> d_x = (p1 -> d_y * p2 -> d_z) - (p1 -> d_z * p2 -> d_y);
	dst -> d_y = (p1 -> d_z * p2 -> d_x) - (p1 -> d_x * p2 -> d_z);
	dst -> d_z = (p1 -> d_x * p2 -> d_y) - (p1 -> d_y * p2 -> d_x);
}

/*
 * Return the dot product:	p1 . p2.
 */

	static
	double
dot3 (

const struct Point3 *		p1,
const struct Point3 *		p2
)
{
double		prod;

	prod =    (p1 -> d_x * p2 -> d_x)
		+ (p1 -> d_y * p2 -> d_y)
		+ (p1 -> d_z * p2 -> d_z);

	return (prod);
}

/*
 * Scale the given Point3 to unit magnitude.
 */

	static
	void
unitize3 (

struct Point3 *		dst,
const struct Point3 *	src
)
{
double		mag;

	mag = mag3 (src);

	FATAL_ERROR_IF (mag <= 0.0);

	dst -> d_x	= src -> d_x / mag;
	dst -> d_y	= src -> d_y / mag;
	dst -> d_z	= src -> d_z / mag;
}

/*
 * Return the magnitude of the given Point3.
 */

	static
	double
mag3 (

const struct Point3 *	src
)
{
double		x, y, z, mag2, mag;

	x	= src -> d_x;
	y	= src -> d_y;
	z	= src -> d_z;

	mag2 = (x * x) + (y * y) + (z * z);

	mag = sqrt (mag2);

	return (mag);
}
