/**********************************************************************

	$Id: plot_sthgp_sec_cd.c,v 1.2 2025/04/13 18:44:20 warme Exp $

	File:	plot_sthgp_sec_cd.c
	Rev:	a-2
	Date:	04/13/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Generate plot data for CD of STHGP subtours.

***********************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.
	a-2:	04/13/2025	warme
		: Generate a separate output file for -r.

***********************************************************************/

#include "fatal.h"
#include <gmp.h>
#include "gmpmisc.h"
#include "logic.h"
#include <math.h>
#include <mpfr.h>
#include "new.h"
#include "s3itypes.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include "sthgp_indicators.h"
#include <string.h>

int		main (int argc, char ** argv);

char *		me;

/*
 * Local Functions
 */

static int		decode_args (int argc, char ** argv);
static void		usage (void);

static bool		div_n_flag;
static bool		log_flag;
static char *		reflected_file;
static bool		square_flag;
static int		xprec;
static int		yprec;

/*
 * The main program.
 */

	int
main (

int		argc,
char **		argv
)
{
int		k, n;
mpfr_prec_t	prec;
FILE *		rfp;
double		x;
mpq_ptr		CDIST2, p, max;
mpfr_t		FTEMP, FMAX;
char		xbuf [64], ybuf [64];

	me = argv [0];

	n = decode_args (argc, argv);

	/* Note: we only generate reflected plots for n = 1000, which	*/
	/* takes some time to compute.  We now compute all the CD	*/
	/* values once, and generate plot data to two different files	*/
	/* (stdout and the -r foo file) when doing reflected plots.	*/

	/* Compute the Centroid Distance squared for each subtour. */
	CDIST2 = new_Q_vector (n + 1);
	sthgp_all_subtour_cd2 (CDIST2, n);

	prec = 128;
	mpfr_init2 (FTEMP, prec);

	if (reflected_file NE NULL) {
		/* Generate reflected plot to separate output file.	*/
		/* This only produces data up to k = floor(n/2).	*/
		rfp = fopen (reflected_file, "w");
		if (rfp EQ NULL) {
			perror (reflected_file);
			exit (1);
		}
		/* Find largest CD. */
		max = &CDIST2 [2];
		for (k = 3; k < n; k++) {
			p = &CDIST2 [k];
			if (mpq_cmp (p, max) > 0) {
				max = p;
			}
		}
		/* Scale CD to max of 1. */
		mpfr_init2 (FMAX, prec);
		mpfr_set_q (FMAX, max, MPFR_RNDN);
		mpfr_sqrt (FMAX, FMAX, MPFR_RNDN);
		for (k = 2; 2*k < n; k++) {
			x = k;
			x /= n;
			sprintf (xbuf, "%.3g", x);

			/* Reflect. */
			mpfr_set_q (FTEMP, &CDIST2 [n - k], MPFR_RNDN);
			mpfr_sqrt (FTEMP, FTEMP, MPFR_RNDN);
			mpfr_div (FTEMP, FTEMP, FMAX, MPFR_RNDN);
			mpfr_sprintf (ybuf, "%.17Rg", FTEMP);

			fprintf (rfp, "%s\t%s\n", xbuf, ybuf);
		}
		mpfr_clear (FMAX);
		fclose (rfp);
	}

	/* Always generate the "main" CD plot to stdout. */
	for (k = 2; k < n; k++) {
		x = k;
		if (div_n_flag) {
			sprintf (xbuf, "%.*g", xprec, x / n);
		}
		else {
			sprintf (xbuf, "%.17g", x);
		}

		mpfr_set_q (FTEMP, &CDIST2 [k], MPFR_RNDN);
		if (NOT square_flag) {
			mpfr_sqrt (FTEMP, FTEMP, MPFR_RNDN);
		}
		if (log_flag) {
			mpfr_log10 (FTEMP, FTEMP, MPFR_RNDN);
		}
		mpfr_sprintf (ybuf, "%.*Rg", yprec, FTEMP);

		printf ("%s\t%s\n", xbuf, ybuf);
	}

	mpfr_clear (FTEMP);

	free_Q_vector (CDIST2, n + 1);

	sthgp_indicators_shutdown ();

	exit (0);
}

/*
 * Decode command line arguments.
 */

	static
	int
decode_args (

int		argc,
char **		argv
)
{
int		i, n;
char		c;
char *		ap;

	div_n_flag	= FALSE;
	log_flag	= FALSE;
	reflected_file	= NULL;
	square_flag	= FALSE;
	xprec		= 3;
	yprec		= 17;
	n		= -1;
	for (i = 1; i < argc; i++) {
		ap = argv [i];
		c = *ap;
		if (c EQ '-') {
			++ap;
			for (;;) {
				c = *ap++;
				if (c EQ '\0') break;
				switch (c) {
				case 'l':
					log_flag = TRUE;
					break;

				case 'n':
					div_n_flag = TRUE;
					break;

				case 'r':
					if (i >= argc) usage ();
					reflected_file = argv [++i];
					ap = "";
					break;

				case 's':
					square_flag = TRUE;
					break;

				case 'x':
					if (i >= argc) usage ();
					ap = argv [++i];
					xprec = atoi (ap);
					ap = "";
					break;

				case 'y':
					if (i >= argc) usage ();
					ap = argv [++i];
					yprec = atoi (ap);
					ap = "";
					break;

				default:
					usage ();
					break;
				}
			}
		}
		else if ((c >= '0') AND (c <= '9')) {
			n = atoi (ap);
			if (n < 3) {
				fprintf (stdout, "%s: Must have N >= 3.\n", me);
				usage ();
			}
			ap = "";
		}
	}
	if (n < 0) {
		/* Default value of n. */
		n = 10;
	}

	return (n);
}

/*
 * Print command usage and exit.
 */

	static
	void
usage (void)

{
	fprintf (stderr,
		 "Usage:\n"
		 "  %s [-[lns]] [-x p] [-y p] [-r file] [ num_verts ]\n"
		 "\t-l\tDisplay log10() of y value\n"
		 "\t-n\tUse x = k / n so that x in (0,1), allowing\n"
		 "\t\tplots for several n to be combined in one plot.\n"
		 "\t-r file\tAlso write data for reflected plot to file.\n"
		 "\t-s\tDisplay centroid distance squared.\n"
		 "\t-x p\tRound x coord to p decimal places.\n"
		 "\t-y p\tRound y coord to p decimal places.\n",
		 me);

	exit (1);
}
