/**********************************************************************

	$Id: plot_tsp_sec_epr.c,v 1.1 2025/01/26 02:31:52 warme Exp $

	File:	plot_tsp_sec_epr.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Generate plot data for EPR of TSP subtours.

***********************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

***********************************************************************/

#include "fatal.h"
#include <gmp.h>
#include "gmpmisc.h"
#include "logic.h"
#include <math.h>
#include <mpfr.h>
#include "new.h"
#include "s3itypes.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include "tsp_indicators.h"
#include <string.h>

int		main (int argc, char ** argv);

char *		me;

/*
 * Local Functions
 */

static int		decode_args (int argc, char ** argv);
static void		usage (void);

static bool		count_flag;
static bool		div_n_flag;
static bool		log_flag;
static int		xprec;
static int		yprec;

/*
 * The main program.
 */

	int
main (

int		argc,
char **		argv
)
{
int		k, n;
double		x;
mpz_ptr		NXP;
mpz_t		TOTAL_TOURS;
mpq_t		RATIO;
mpfr_t		FTEMP;
char		xbuf [64], ybuf [64];

	me = argv [0];

	n = decode_args (argc, argv);

	/* Compute the number of extreme points incident to each subtour. */
	NXP = new_Z_vector (n + 1);
	for (k = 2; k <= n - 2; k++) {
		tsp_subtour_ep_count (&NXP [k], n, k);
	}

	mpz_init (TOTAL_TOURS);
	mpq_init (RATIO);
	mpfr_init2 (FTEMP, 128);

	tsp_num_extreme_points (TOTAL_TOURS, n);

	for (k = 2; k <= n - 2; k++) {
		x = k;
		if (div_n_flag) {
			sprintf (xbuf, "%.*g", xprec, x / n);
		}
		else {
			sprintf (xbuf, "%.17g", x);
		}

		if (count_flag) {
			/* Plot raw counts */
			mpfr_set_z (FTEMP, &NXP [k], MPFR_RNDN);
			if (log_flag) {
				mpfr_log10 (FTEMP, FTEMP, MPFR_RNDN);
			}
		}
		else {
			/* Plot EPR */
			mpz_set (mpq_numref (RATIO), &NXP [k]);
			mpz_set (mpq_denref (RATIO), TOTAL_TOURS);
			mpq_canonicalize (RATIO);
			mpfr_set_q (FTEMP, RATIO, MPFR_RNDN);
			if (log_flag) {
				mpfr_log10 (FTEMP, FTEMP, MPFR_RNDN);
			}
		}
		mpfr_sprintf (ybuf, "%.*Rg", yprec, FTEMP);

		printf ("%s\t%s\n", xbuf, ybuf);
	}

	mpfr_clear (FTEMP);
	mpq_clear (RATIO);
	mpz_clear (TOTAL_TOURS);

	free_Z_vector (NXP, n + 1);

	tsp_indicators_shutdown ();

	exit (0);
}

/*
 * Decode command line arguments.
 */

	static
	int
decode_args (

int		argc,
char **		argv
)
{
int		i, n;
char		c;
char *		ap;

	count_flag	= FALSE;
	div_n_flag	= FALSE;
	log_flag	= FALSE;
	xprec		= 3;
	yprec		= 17;
	n		= -1;
	for (i = 1; i < argc; i++) {
		ap = argv [i];
		c = *ap;
		if (c EQ '-') {
			++ap;
			for (;;) {
				c = *ap++;
				if (c EQ '\0') break;
				switch (c) {
				case 'c':
					count_flag = TRUE;
					break;

				case 'l':
					log_flag = TRUE;
					break;

				case 'n':
					div_n_flag = TRUE;
					break;

				case 'x':
					if (i >= argc) usage ();
					ap = argv [++i];
					xprec = atoi (ap);
					ap = "";
					break;

				case 'y':
					if (i >= argc) usage ();
					ap = argv [++i];
					yprec = atoi (ap);
					ap = "";
					break;

				default:
					usage ();
					break;
				}
			}
		}
		else if ((c >= '0') AND (c <= '9')) {
			n = atoi (ap);
			if (n < 3) {
				fprintf (stdout, "%s: Must have N >= 3.\n", me);
				usage ();
			}
			ap = "";
		}
	}
	if (n < 0) {
		/* Default value of n. */
		n = 10;
	}

	return (n);
}

/*
 * Print command usage and exit.
 */

	static
	void
usage (void)

{
	fprintf (stderr,
		 "Usage: %s [-[cln]] [ num_verts ]\n"
		 "\t-c\tDisplay counts instead of ratios\n"
		 "\t-l\tDisplay log10() of y value\n"
		 "\t-n\tUse x = k / n so that x in (0,1), allowing\n"
		 "\t\tplots for several n to be combined in one plot.\n"
		 "\t-x p\tRound x coord to p decimal places.\n"
		 "\t-y p\tRound y coord to p decimal places.\n",
		 me);

	exit (1);
}
