/**********************************************************************

	$Id: sthgp_sec_cd2_dx_minmax.c,v 1.1 2025/01/26 02:31:52 warme Exp $

	File:	sthgp_sec_cd2_dx_minmax.c
	Rev:	a-1
	Date:	01/25/2025

	Copyright (c) 2012, 2025 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Plot or test DX_e values and sums of (DX_e)^2 values
	for the STHGP(n) subtour Centroid Distance.

***********************************************************************

	Modification Log:

	a-1:	01/25/2025	warme
		: Cleanup for initial release.

***********************************************************************/

#include "fatal.h"
#include <gmp.h>
#include "logic.h"
#include <math.h>
#include "s3itypes.h"
#include <stdlib.h>
#include <stdio.h>
#include "sthgp_indicators.h"
#include "sthgp_sec_cd2_dx.h"
#include <string.h>


/*
 * Global Routines
 */

int		main (int, char **);

char *		me;


/*
 * Local Routines
 */

static void		compute_min_max_dx (int n);
static void		usage (void);

/*
 * The main routine.
 *
 * Given the number of vertices N, it computes the centroid distance
 * for the K-vertex subtour for each 2 <= K <= N-1.
 */

	int
main (

int		argc,
char **		argv
)
{
int		n;

	me = argv [0];

	if (argc NE 2) {
		usage ();
	}

	n = atoi (argv [1]);
	if (n < 3) {
		fprintf (stderr, "%s: must have at least 3 vertices.\n", me);
		exit (1);
	}

	compute_min_max_dx (n);

	sthgp_indicators_shutdown ();

	exit (0);
}

/*
 * Display command usage and exit.
 */

	static
	void
usage (void)

{
	fprintf (stderr, "Usage: %s NVerts\n", me);
	exit (1);
}

/*
 * Compute the smallest and largest DX(n,k,p,q) magnitudes.
 */

	static
	void
compute_min_max_dx (

int		n
)
{
int		k, p, q, esize, pval, pqval;
int		min_k, min_p, min_q;
int		max_k, max_p, max_q;
bool		have_min_max;
mpq_t		TAU_VAL, VAL, MIN_VAL, MAX_VAL;
mpz_t		A, B, TMP1, TMP2;

	FATAL_ERROR_IF (n < 3);

	mpq_init (TAU_VAL);
	mpq_init (VAL);
	mpq_init (MIN_VAL);
	mpq_init (MAX_VAL);

	mpz_init (A);
	mpz_init (B);
	mpz_init (TMP1);
	mpz_init (TMP2);

	min_k = min_p = min_q = 0;
	max_k = max_p = max_q = 0;
	have_min_max = FALSE;

	sthgp_subtour_cd2_Alpha (A, n);

	for (k = 2; k < n; k++) {
		sthgp_subtour_cd2_dx_TAU (TAU_VAL, n, k);
		sthgp_subtour_cd2_Beta (B, n, k);

		for (esize = 2; esize <= n; esize++) {
			for (p = 0; p <= esize; p++) {
				q = esize - p;
				pval = p - 1;
				if (pval < 0) {
					pval = 0;
				}
				pqval = p + q - 1;
				if (pqval < 0) {
					pqval = 0;
				}

				mpz_mul_ui (TMP1, A, pval);
				mpz_mul_ui (TMP2, B, pqval);
				mpz_sub (TMP1, TMP1, TMP2);

				mpq_set (VAL, TAU_VAL);
				mpz_mul (mpq_numref (VAL),
					 mpq_numref (VAL),
					 TMP1);
				mpq_canonicalize (VAL);
				mpq_abs (VAL, VAL);

				if (NOT have_min_max) {
					min_k = max_k = k;
					min_p = max_p = p;
					min_q = max_q = q;
					mpq_set (MIN_VAL, VAL);
					mpq_set (MAX_VAL, VAL);
					have_min_max = TRUE;
				}
				else {
					if (mpq_cmp (VAL, MIN_VAL) < 0) {
						min_k = k;
						min_p = p;
						min_q = q;
						mpq_set (MIN_VAL, VAL);
					}
					if (mpq_cmp (VAL, MAX_VAL) > 0) {
						max_k = k;
						max_p = p;
						max_q = q;
						mpq_set (MAX_VAL, VAL);
					}
				}
			}
		}
	}

	printf ("Min at DX(%d, %d, %d, %d) = %.17g\n",
		n, min_k, min_p, min_q, mpq_get_d (MIN_VAL));

	printf ("Max at DX(%d, %d, %d, %d) = %.17g\n",
		n, max_k, max_p, max_q, mpq_get_d (MAX_VAL));

	mpz_clear (TMP2);
	mpz_clear (TMP1);
	mpz_clear (B);
	mpz_clear (A);

	mpq_clear (MAX_VAL);
	mpq_clear (MIN_VAL);
	mpq_clear (VAL);
	mpq_clear (TAU_VAL);
}
